//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <BrashCommands.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/stringexpressionparser.h>
#include <cxxtls/ArithmeticExpressionParser.h>
#include <cxxtls/getopt.h>
#include <fstream>


using namespace std;
using namespace cxxtls;
using namespace ScriptFunctions;

static string colon(":");
static string nl("\r\n");
static string comma(",");
static string space(" ");
static string indent("    ");
static string equalSign("=");
static string is(" is ");
static string in(" in ");
static string builtin("builtin");
static string alias("alias");
static string unknown("unknown");
static string apostrophe("'");
static string let("let");
static string test("test");
static string errorKeyword("Error ");  // includes trailing space
static string elipses("...");


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

enum JoinType
{
   JOIN_DIRECT,         // when joining continued lines, use no separators
   JOIN_NEWLINE,        // use a newline separator
   JOIN_SEMICOLON       // use a semicolon separator -- actually " ; "
};

static bool skipOverDoubleQuotes(string::const_iterator &f, string::const_iterator &l, JoinType &joinType);
static bool skipOverBackTicks(string::const_iterator &f, string::const_iterator &l, JoinType &joinType);
static bool skipOverDollarParenSubshell(string::const_iterator &f, string::const_iterator &l, JoinType &joinType);
static bool skipOverDollarCurlyVar(string::const_iterator &f, string::const_iterator &l, JoinType &joinType);



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool skipOverSingleQuotes(string::const_iterator &f, string::const_iterator &l, JoinType &joinType)
   // move f towards l and until apostrophes have been matched.
   //
   // return true if f reaches l without it
   //
   // leave f after the trailing apostrophe
{
    JoinType tmpJoin = joinType;

    if(f == l)
       return false;  // on startup, who cares

    if(*f != '\'')
       return false;  // bad invocation, just ignore it

    ++f;

    joinType = JOIN_NEWLINE;

    while(f != l && *f != '\'')
       ++f;

    if(f == l)
       return true;  // bad syntax

    ++f;

    joinType = tmpJoin;
    return false; // s'all good

}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool skipOverBackTicks(string::const_iterator &f, string::const_iterator &l, JoinType &joinType)
   // move f towards l and until back-ticks have been matched.
   //
   // return true if f reaches l without it
   //
   // leave f after the trailing tick
{

    JoinType tmpJoin = joinType;

    if(f == l)
       return false;  // on startup, who cares

    if(*f != '`')
       return false;  // bad invocation, just ignore it

    joinType = JOIN_SEMICOLON;  // separating statements

    ++f;

    while(f != l && *f != '`')
       ++f;

    if(f == l)
       return true;  // bad syntax

    ++f;

    joinType = tmpJoin;

    return false; // s'all good

}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool skipOverDoubleQuotes(string::const_iterator &f, string::const_iterator &l, JoinType &joinType)
   // skip over a double quoted string, but handling nested back-tick subshells and 
   // also handling $() subshells
{
   JoinType tmpJoin = joinType;

   string::const_iterator orgF = f;
   
   if(f == l)
      return false;

   if(*f != '"')
      return false;

   ++f; 

   joinType = JOIN_NEWLINE;

   while(f != l && *f != '"')
   {
      char c = *f++;

      if(c == '\\')
      {
         if(f == l)
            return true; // unmatched "

         ++f;  // skip escaped value
      }
      else
      if(c == '`')
      {
         if(skipOverBackTicks(f,l, joinType))
            return true;

         // otherwise, s'all good and trailing back tick has been removed
      }
      else
      if(c == '$' && f != l && *f == '(')
      {
         --f;
         if(skipOverDollarParenSubshell(f,l, joinType))
            return true;

         // otherwise, s'all good and trailing ) has been removed
      }
      else
      if(c == '$' && f != l && *f == '{')
      {
         --f;
         if(skipOverDollarCurlyVar(f,l, joinType))
            return true;

         // otherwise, s'all good and trailing ) has been removed
      }

   }

   if(f == l)
      return true;

   ++f;  // skip trailing "

   joinType = tmpJoin;

   return false;

}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool skipOverDollarParenSubshell(string::const_iterator &f, string::const_iterator &l, JoinType &joinType)
   // move f towards l until $() is finished.  Leave f after the ).  Return true if f== l and
   // we have not matched ()
{

    JoinType tmpJoin = joinType;
    
    if(f == l)
       return false;

    if(*f != '$')
       return false;

    // f does point to $

    ++f;

    if(f == l || *f != '(')
       return false;

    joinType = JOIN_SEMICOLON;  // command separator

    // at this point, we know that the stream containined $(, so look for trailing ).

    ++f;  // skip leading (

    while(f != l && *f != ')')
    {
       char c = *f++;

       if(c == '\'')
       {
          --f;
          if(skipOverSingleQuotes(f,l, joinType))
             return true;

          // otherwise, f is past trailing apostrophe
       }
       else
       if(c == '\\')
       {
          if(f == l)
             return true;

          ++f;  // skip escaped value
       }
       else
       if(c == '`')
       {
           --f;
           if(skipOverBackTicks(f,l, joinType))
              return true;

           // trailing tick has already been skipped
       }
       else
       if (c == '"')
       {
           --f;
           if (skipOverDoubleQuotes(f,l,joinType))
               return true;
       }
       else
       if(c == '$' && f != l && *f == '(')
       {
          --f;
          if(skipOverDollarParenSubshell(f,l, joinType))
             return true;
          // trailing ) has been removed
       }
       else
       if(c == '$' && f != l && *f == '{')
       {
          --f;
          if(skipOverDollarCurlyVar(f,l, joinType))
             return true;
          // trailing ) has been removed
       }
    }

    if(f == l)
       return true;

   ++f;  // skip trailing )

   joinType = tmpJoin;

   return false;

}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool skipOverDollarCurlyVar(string::const_iterator &f, string::const_iterator &l, JoinType &joinType)
{
    JoinType tmpJoin = joinType;
    
    if(f == l)
       return false;

    if(*f != '$')
       return false;

    ++f;

    if(f == l)
       return false;

    if(*f != '{')
       return false;

    joinType = JOIN_DIRECT;  // use no separator

    ++f;

    while(f != l && *f != '}')
    {
       char c = *f++;

       if(c == '\\')
       {
          if(f == l)
             return true;

          ++f;
       }
       else
       if( c == '$' && f != l && *f == '{')
       {
          --f;
          if(skipOverDollarCurlyVar(f,l, joinType))
             return true;
       }

    }

    if(f == l)
       return true;

    ++f;  // skip trailing }


    joinType = tmpJoin;

    return false;

}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool hasOpenQuotes(string::const_iterator &f, string::const_iterator &l, JoinType &joinType)
    // move f towards l checking on whether or not there is an unclosed pair of quotes if f reach l.
    //
    // return true to indicate that there is an unclosed quote block
{
    while(f != l)
    {
       char c = *f++;

       if(c == '\\')
       {
          if(f == l)
             return true;

          ++f;  // skip escaped value
       }
       else
       if(c == '$' && f != l && *f == '(')
       {
          --f;
          if(skipOverDollarParenSubshell(f,l, joinType))
             return true;

          // trailing ) is already gone
       }
       else
       if(c == '$' && f != l && *f == '{')
       {
          --f;
          if(skipOverDollarCurlyVar(f,l, joinType))
             return true;

          // trailing ) is already gone
       }
       else
       if(c == '\'')
       {
          --f;
          if(skipOverSingleQuotes(f,l, joinType))
             return true;

          // trailing apostrophe is gone
       }
       else
       if(c == '`')
       {
          --f;
          if(skipOverBackTicks(f,l, joinType))
             return true;

          // trailing tick is gone
       }
       else
       if (c == '"')
       {
          --f;
          if (skipOverDoubleQuotes(f,l, joinType))
             return true;
          // trailing double quote is already gone
       }
       else
       if(c == '#')
       {
          while( f != l )
            ++f;
       }

    }

    return false;  // nothing to check for here
}



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


static bool hasBrashContinuationFlag(ScriptInputLine const &line, bool &keepNewline, JoinType &joinType)
    /// return true if a given command.text_ string has the expected
    /// line continuation character sequence -- or if there is an
    /// unclosed double quote sequence left over from this or some
    /// previous line
{
    keepNewline = false; // for \\ the answer is no

    size_t charCount = line.text_.size();

    if(
        (   // last char on the line is backslash
            charCount != 0
            && line.text_[charCount - 1] == '\\'
            && (charCount == 1 || line.text_[charCount - 2] != '\\')
        )
        ||   
        ( // last char sequence is \ carriateReturn newline
#ifdef _MSC_VER
            charCount >= 3
       &&   line.text_[charCount-1] == '\n'
       &&   ((     line.text_[charCount-2] == '\r'
              &&   line.text_[charCount-3] == '\\'
             )  
             || line.text_[charCount - 2] == '\\'
            )
#else
            line.text_.size() >= 2
       &&   line.text_[line.text_.size()-2] == '\\'
#endif
      )
      )
    {
       joinType = JOIN_DIRECT;

       return true;
    }

    keepNewline = true; // we may return both false and keepNewline = true.  KeepNewline will be ignore in this case


    // yuck!  now we need to make sure that the number of ', `, and " characters are matched on a line

    string::const_iterator f = line.text_.begin(),
                           l = line.text_.end();

    return hasOpenQuotes(f,l, joinType);

}


bool handleLineContinuations(ScriptLineStream  &stream, ScriptInputLine &command, bool wholeFileAsLine)
{
    auto_ptr< AutoPoppingPushPrompt > pushPromptObject;

    if(wholeFileAsLine)
    {
        return stream.readFile(command);   // read the whole file as a single string
    }


    if(stream.readLine(command))
    {
        bool keepNewline = false;

        JoinType joinType = JOIN_DIRECT;

        while( hasBrashContinuationFlag(command, keepNewline, joinType) )
        {
            if(!pushPromptObject.get())
            {
               pushPromptObject.reset( new AutoPoppingPushPrompt(stream, stream.getPrompt() + "> ") );
            }

            size_t trailingBackSlash = size_t(-1);

            if (!keepNewline)
               trailingBackSlash = command.text_.find_last_of('\\');  // line ended with backslash not " ... or '....

            if(trailingBackSlash < command.text_.size())
                command.text_.erase(trailingBackSlash, command.text_.size());

            bool foundMore = false;

            ScriptInputLine more;

            bool foundMoreLines = stream.readLine(more);

            if(!foundMoreLines)
            {
               command.text_ += '\n';  // put the trailing \n back on the line
               break;
            }


            if(joinType == JOIN_NEWLINE)
               command.text_ += '\n';
            else
            if(joinType == JOIN_SEMICOLON)
               command.text_ += " ; ";

            command.text_ += more.text_;  

        }

        return true;
    }


    return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void skipEmptyStatements(string::const_iterator           &f, 
                         string::const_iterator           &l,
                         ScriptLineStream                 &stream,
                         ScriptInputLine                  &command
                        )
   /// skip past blank lines, single semicolons, and begining of line comments 
{

   for(;;)
   {
       // skip blanks

       StringExpr::skipSeparators(f,l);

       if(f == l)
       {
           // handle end of line

          
           if(!handleLineContinuations(stream, command))
           {
               break;
           }
       
           f = command.text_.begin();
           l = command.text_.end();
       
       }

       if(*f == ';')
       {
          string::const_iterator next = f; ++next;

          if(next == l)
             continue;   

          if(*next == ';')
             break;  // f points to ;; which is not what we are looking for

          ++f;  // skip the end of line and keep scanning
       }
       else
       if(*f == '#')
       {
           while(f != l && *f != '\n')
              ++f;

           // let the outer loop remove the \n
       }
       else
       {
           break;
       }

   }
    
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

string parseFirstWord(string::const_iterator &f, string::const_iterator const &l)
   // parse the first word of interest to the brash language. 
   //
   // words are things like alphanumeric strings but they can also include [], (), {}, etc
   //
   // STOP:  be extremely careful about changing this function.  It is not meant as a general purpose
   //        parsing tool.  Rather, it exists to parse the first word on a command line and as such
   //        it is incomplete in some areas and does wierd things in other areas.  Sometimes, it refuses to
   //        parse the first character and leaves an empty return string -- with the assumption that
   //        some other piece of code will figure out what to do with it.
{
   StringExpr::skipSeparators(f,l);

   string::const_iterator startF = f;

   string rv;

   for(;;)
   {
      char c = 0;

      if(f == l)
         break;

      c = *f;

      bool ok = (isalnum(c) || 
                c == '_'    || 
                c == '['    || 
                c == '('    || 
                c == ')'    || 
                c == '{'    || 
                c == '}'    || 
                c == '.'    ||
                c == '/'    || 
                c == ':'    || 
                c == '\\'   || 
                c == '$'    || 
                c == '\''   || 
                c == '"'
               );

      if(!ok)
         break;

      rv += *f++;
   }

   if(f == l)
      return rv;

   static string emptystring;

   if(!isspace(*f) && f == startF)
      return emptystring;

   return rv;
    
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

string offendingCommandFragment(string::const_iterator f, ScriptInputLine const &command)
    /// construct a string that describes where in a given source a syntax error has occurred.
{
    size_t offset = f - command.text_.begin();

    static string  badloc = "unknown location in:  " ;

    string word;

    if (offset < command.text_.size())
    {
        word = command.text_.substr(offset, command.text_.size());
    }
    else
    {
        if(command.text_.size() > 40)
        {
            word = command.text_.substr(0, 40) + elipses;
        }
        else
            word = command.text_;
    }

    return command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + colon + space + apostrophe + word + apostrophe;

}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

string ErrorMessageWithLocation(string const &message, 
                                string::const_iterator f, 
                                ScriptInputLine const &command
                               )
{
    return errorKeyword + offendingCommandFragment(f, command) + colon + space + message;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static
cxxtls::StringExpression const*
findInlineInputRedirection(Statement *statement)
{
   BrashCommandLineStatement *cmdLineStatement = dynamic_cast<BrashCommandLineStatement*>(statement);

   if(cmdLineStatement)
   {
       return cmdLineStatement->command_->findInlineInputRedirection();
   }
   else
   {
       // handle other types
   }

   return 0;
  
}



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseCommandLine(string::const_iterator           &f, 
                             string::const_iterator           &l,
                             ScriptLineStream                 &stream,
                             StatementListStatement           &output,
                             ScriptThread::ManagedFileHandles &messageFiles,
                             ScriptInputLine                  &command,   // over which f and l iterate
                             ScriptThread::ThreadContext      *parsingContext
                            )
    /// Parse a normal command line out of the input text range, (f,l].  The iterators run over the parameter
    /// member, command.text_.  This function may repopulate command.text_ from stream if the text on the line
    /// implies some for of continuation, such as if statements, while statements, for statements, etc.
    /// The parsed command is pushed_back into the output statement list.  
    ///
    /// If the command terminator is &, then the caller is expected to change it such that it becomes a 
    /// separate stand-alone command with its own environment.
{
   // f and l iterate over command

   auto_ptr<BrashCommandLine::ARGUMENTS_LIST  >       arguments    (new BrashCommandLine::ARGUMENTS_LIST);
   auto_ptr<BrashCommandLine::REDIRECTION_LIST>       redirections (new BrashCommandLine::REDIRECTION_LIST);


   typedef list< pair<string,StringExpression*> >     VAR_SUPERSEEDS;

   auto_ptr<list< pair<string,StringExpression*> > >  superseeds   (new VAR_SUPERSEEDS);

   // first check to see if this command is nothing more than a variable assignment.  The alternative to 
   // that is a collection of superseeded variables then a command

   string startFilename = stream.filename();        // for CXXTLS_DEBUG, below
   int    startLineNumber = stream.lineNumber();  
   string::const_iterator startF = f;


   bool expressionCommand = false;
   string exprType("?");
   bool discoverRedirections = true;

   StringExpr::skipSeparators(f,l);


   // check for builtin expression statements

   if(f != l && *f == '[' || *f == '(')
   {

      // determine if we have either of these commands:
      //
      //   (( ... ))  // we should never get here under these circumstance
      //   [ ... ]
      //   [[ ... ]]

      expressionCommand = true;

      exprType.resize(0);
      exprType += *f;

      string::const_iterator next = f;

      ++next;

      if(next != l)
      {
         if(*next == *f)
         {
            exprType += *next;

            if(*next == '[')
                discoverRedirections = false;
         }

      }

   }

   // check for variable superseeds occuring before the command name


   bool assignmentOnly = false;

   while(!expressionCommand)
   {  // collect variable superseeds

      string::const_iterator parsingStart = f;

      string variableName = parseFirstWord(f,l);

      if(variableName.empty())
      {
         break;
      }

       
      if(f == l || (*f != '='))
      {
         // we have the first word of a command, not a variable assignment

         f = parsingStart;
         break;
      }
      else
      {
          ++f;
      }

      // if we get here, we have
      //
      //   x=...

      
      parsingStart = f;

      string filename = stream.filename();
      int    line = stream.lineNumber();

      auto_ptr<StringExpression> e(StringExpr::parse(f,l));

      if (e.get())
      {
          // keep together for exception safety
          superseeds->push_back( pair<std::string, cxxtls::StringExpression*>(variableName, e.get() )  );
          e.release();
      }
      else
      {
          string errorMessage = filename + colon + lexical_cast<string>(line)+colon + space + ":  Invalid string syntax\r\n";
          messageFiles.writeBytes(2, errorMessage);
          return false;
      }

      // now check for this:    "v=`some command`\n if [ stuff ] ..."

      while(f != l && isspace(*f) && *f != '\n')
      {
         ++f;
      }

      // allow a comment at the end of the line

      if (f != l && *f == '#')
      {
          while (f != l && *f != '\n')
             ++f;  
      }

      if(f != l && *f == '\n')
      {
         // if we get here, it is because there was a variable assignment in a file, rather than in the console
         // and the variable assignment should not get stuck to the command that follows it as a superseed unless
         // the command is on the same line

         assignmentOnly = true;
         break;  
      }


   }  // collect variable superseeds
   
   // at this point, we now have variable superseeds populated with 0 or more variable assignments.

   StringExpr::skipSeparators(f,l);

   if( assignmentOnly || f == l || *f == ';' || *f == '&' || *f == '|')
   {
      // we are at the end of the statement

      if(superseeds->size() == 0)
      {
         messageFiles.writeBytes(2, command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + string(": Invalid command line syntax\r\n"));
         return false;
      }

      if(superseeds->size() > 1 )
      {
         messageFiles.writeBytes(2, command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + string(": Invalid variable assignment syntax\r\n"));
         return false;
      }

      {
          // keep all this together for exception safety.

          pair<string, cxxtls::StringExpression*> back = superseeds->back();  // get the one superseed off the list
          
          auto_ptr<cxxtls::StringExpression> valueExpression(back.second);  // don't let the allocated memory hang around, unprotected
          
          superseeds->pop_back(); // discard that only superseed item in the list 
          
          {
             output.statements_->push_back(ThreadSafeHandle<Statement>(new VariableAssignmentStatement(back.first, valueExpression.get())));
             valueExpression.release();
          }
      }

      return true;  // we have nothing but a variable assignment in this statement.


   }

   // now parse the command name and parameters

   bool foundOne = false;

   for(;;)
   {  // parse all command line parameters
       
       string::const_iterator  savedF = f;  // used only when hitting statement terminators.


       StringExpr::skipSeparators(f,l);

       if(f == l)
          break;


       if(*f == ';' || *f == '&' || *f == '|')
       {
           // new statement has arived

           f  = savedF;
           break;

       }


       if(*f == '#')
       {
           // we have a comment to end of line

           f  = l;
           break;

       }

       savedF = f;  // used only when hitting statement terminators.

       auto_ptr<cxxtls::StringExpression>   expression( cxxtls::StringExpr::parseStrExpr(f,l, discoverRedirections) );

       if (expression.get() == 0)
       {
           if (savedF != l  )
           {
               char buffer[2];   

               buffer[0] = *savedF;
               buffer[1] = 0;

               messageFiles.writeBytes(2, command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + string(": invalid token after '") + buffer + string("'\r\n") );
               return false;
           }

           break;
       }


       if (expressionCommand && exprType == "[[")
       {
           string word;

           if (expression->isSimple(word) && word == "]]")
           {
                // end of statement found

               StringExpr::skipSeparators(f, l);

               if (f != l)
               {
                   if (*f != '#' && *f != ';' && *f != '&')
                   {
                       messageFiles.writeBytes(2, command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + string(": text found after ]] \r\n"));
                       return false;
                   }
               }

               break;
           }
       }

       if(expression->isRedir())
       {
           string value = expression->stringify();

           // here are the expected values
           //
           //     >fred
           //     <tom
           //     >>fred
           //     <<tom
           //     2>fred
           //     9>>fred
           //    43>>&99  


           if(!value.empty())
           {
              char c = value[0];

              size_t bracketOffset = value.find_first_of("<>");

              if(bracketOffset >= value.size())
              {
                  messageFiles.writeBytes(2, command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + colon + string("Error Invalid redirection\r\n"));
                  return false;
              }

              int fileNo = -1;
              int otherFile = -1;
              bool input = false;
              bool append = false;

              if(value[bracketOffset] == '>')
                 fileNo = 1;
              else
              {
                 fileNo = 0;
                 input = true;
              }

              
              if(isdigit(c))
              {
                 // numbered redirection

                 fileNo = lexical_cast<int>(value);  // hopefully this won't throw an exception due to the <> after the number

              }

              char const *p = &value[bracketOffset];
              char const *e = &value[0] + value.size();


              if(p != e)
              {
                  // check for open append and for &otherfile specification

                  ++p;

                  if(p != e)
                  {
                     if(*p == c)
                     {
                         ++p;
                         append = true;
                     }

                     if(p == e)
                     {
                          messageFiles.writeBytes(2, command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + colon + string("Error Invalid redirection\r\n"));
                          return false;
                     }

                     if(*p == '&')
                     {
                        ++p;

                        if(p == e)
                        {
                          messageFiles.writeBytes(2, command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + colon + string("Error Invalid redirection\r\n"));
                          return false;
                        }

                        otherFile = 0;

                        while(p != e && isdigit(*p))
                        {
                           otherFile *= 10;
                           otherFile += *p++ - '0';
                        }

                        // otherfile now has the correct number in it.

                     } // &otherfile

                  } // not out of data

              } // checking for append and for &otherfile

              RedirType redirType(fileNo, input, append, otherFile);

              {
                  // for exception saftey, do not separate these two statements
                  redirections->push_back( pair<RedirType, cxxtls::StringExpression*>( redirType, expression.get() ) );
                  expression.release();
              }

           }  // redirection processing
           else
           {
               // we have a redirection but the string value is empty?

               messageFiles.writeBytes(2, command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + colon + string("Error Invalid redirection\r\n"));
               return false;
           }
       }
       else
       if(!expressionCommand && expression->isSpecial())
       {
          // some kind of statement terminator

          f = savedF;  // unparse the terminator
          break;
       }
       else
       {
           string aliasableName;

           if(!foundOne && expression.get() && expression->isSimple(aliasableName))
           {
               // this code works only on the first word of a command -- basically, it looks up the word to see if is an alias
               // and the substitutes the alias contents for the word.

               StringExpr::ExecutionContext::AliasMap::const_iterator where = parsingContext->variables().aliases().find(aliasableName);

               if(where ==  parsingContext->variables().aliases().end())
               {
                   // for exception safety, do not separate these two statements

                   if (exprType == "[[")
                   {
                       expression.release();  // discard [[ and replace ith with "test"

                       string::const_iterator testBeg = test.begin(), 
                                              testEnd = test.end();

                       expression.reset( cxxtls::StringExpr::parseStrExpr(testBeg, testEnd, discoverRedirections) );

                   }

                   arguments->push_back(expression.get());
                   expression.release();
               }
               else
               {
                   CXXTLS_FOREACH(string const &cur, where->second)
                   {
                      auto_ptr<StringExpression> exp(new StringExpression() );

                      auto_ptr<StringExpr::LiteralNode>  stringNode(new StringExpr::LiteralNode(cur));

                      {
                          // keep together for string safety

                          exp->adoptNode( stringNode.get() );
                          stringNode.release();
                      }

                      {
                          // keep together for exception safety

                          arguments->push_back(exp.get());
                          exp.release();
                      }
                       
                   }

                   expression.release();   // the name that became aliased
               }
           }
           else
           {
               // for exception safety, do not separate these two statements
               arguments->push_back(expression.get());
               expression.release();
           }

           foundOne = true;
       } 

       if (expressionCommand)
       {
           while (f != l && *f != '\n' && isspace(*f))
              ++f;   // skip till end of line

           if (f != l && *f == '\n')
              break; // we are at the end of the [] pair
       }
   } // parse all command line parameters
    
   if(foundOne)
   {
       auto_ptr<BrashCommandLine>  commandLine;

       {
          // for exception saftey, do not separate the following lines

          auto_ptr<BrashCommandLine::SUPERSEEDS> tmpSuperseeds(new BrashCommandLine::SUPERSEEDS);

          typedef pair<string, StringExpression*> SUPERSEED;

          CXXTLS_FOREACH(SUPERSEED &cur, *superseeds)
          {
             try
             {
                 tmpSuperseeds->insert(  pair<string, StringExpression*>(cur.first, cur.second));
             }
             catch(...)
             {
                delete cur.second;
             }
          }
          
          commandLine.reset(new BrashCommandLine(arguments.get(), redirections.get(), tmpSuperseeds.get(), startFilename, startLineNumber ) );
          arguments.release();
          redirections.release();
          tmpSuperseeds.release();
       }

       {
          // for exception saftey, do not separate the following lines

          bool maybeBuiltin = true;   // by default, this command might be a builtin.  Might not, but it might.

          if(! commandLine->arguments_->empty())
          {
              string commandName;

              if((*commandLine->arguments_->begin())->isSimple(commandName))
              {
                 StringExpr::ExecutionVariables::FunctionInfo funcBody = parsingContext->variables().findFunction(commandName);

                 if(funcBody.statement_)
                 {
                    maybeBuiltin = false;   // there is a function by the same name as this function, so let functions superseed builtins
                 }
              }
          }

          CXXTLS_DEBUG(startFilename << ":" << startLineNumber << ":  Creating Statement  " << *commandLine.get());
          
          output.statements_->push_back(
                                          ThreadSafeHandle<Statement>(
                                                                        new BrashCommandLineStatement(commandLine.get(), maybeBuiltin)
                                                                     )
                                       );
          commandLine.release();
       }
   }

   return foundOne;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseFunction      ( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               );


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


      ParseStatementReturnCode
                parseStatement(
                               string::const_iterator           &f,
                               string::const_iterator           &l,
                               ScriptLineStream                 &stream,
                               StatementListStatement           &output,
                               ScriptThread::ManagedFileHandles &messageFiles,
                               ScriptInputLine                  &command,
                               EndOfStatementListDetector const &endDetector,
                               ScriptThread::ThreadContext      *parsingContext
                              );

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// parse the FC command and execute it at parse time...

static bool parseFcStatement    ( string::const_iterator           &f, 
                                   string::const_iterator           &l,
                                   ScriptLineStream                 &stream,
                                   StatementListStatement           &output,
                                   ScriptThread::ManagedFileHandles &messageFiles,
                                   ScriptInputLine                  &command,   // over which f and l iterate
                                   ScriptThread::ThreadContext      *parsingContext
                                 );

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


      static bool
                parseArithmeticStatement(
                                         string::const_iterator           &f,
                                         string::const_iterator           &l,
                                         ScriptLineStream                 &stream,
                                         StatementListStatement           &output,
                                         ScriptThread::ManagedFileHandles &messageFiles,
                                         ScriptInputLine                  &command
                                        );
                // note:  this parser handles "let" and "(( ... ))" statements.
                //
                //        let statements are implemented as a builtin command which evaluates all its arguments
                //        and treats each as an assignment or a comparison.

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseWhileStatement( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               )
{
   // while statements have the following syntax:
   //
   //   while statement ; do   statementList ; done  1>&2 2>nul 0<nul
   //
   // note that ; and \n are synonymous in this example.
   //
   // Until statements are also handed by this function.  They look the same as while statements but the
   // logic of the condition is inverted.

   AutoPoppingPushPrompt  prompt(stream, stream.getPrompt() + "-while> " );  // will auto-restore on return


   string parsingFilename = command.filename_;
   int    parsingStartLine= command.lineNumber_;

   StringExpr::skipSeparators(f,l);

   string::const_iterator beforeCommandName = f;

   string firstWord = parseFirstWord(f,l);

   bool   isUntilStatement = false;

   if(firstWord == "until")
   {
      isUntilStatement = true;
   }
   else
   if(firstWord != "while")
   {
      messageFiles.writeBytes(2, ErrorMessageWithLocation("while-statement-syntax:  missing 'while' keyword", beforeCommandName, command) + nl);
      return false;
   }

   StringExpr::skipSeparators(f,l);

   if(f==l)
   {
      messageFiles.writeBytes(2, ErrorMessageWithLocation("while-statement-syntax:  missing condition statement", f, command) + nl );
      return false;
   }

   ThreadSafeHandle<Statement> conditionStatement;

   {
       // parse the condition statement

       StatementListStatement  conditionTmp(new StatementList);   
       
       // we are going to parse into this and then extract the one statement and put it in conditionStatement.
     
       EndOfStatementListDetector endDetector;

       string::const_iterator beforeCondition = f;

       ParseStatementReturnCode rc = parseStatement(f, l, stream, conditionTmp, messageFiles, command, endDetector, parsingContext);

       if(PSRC_BAD == rc)
       {
           return false;  
       }

       // now move the thing just parsed from the condition temporary list into the conditionStatement variable
     
       conditionStatement = conditionTmp.statements_->statements().back();

       conditionTmp.statements_->pop_back(); 

       if(isUntilStatement)
       {
          // until is like while except that the condition logic is reversed

          ThreadSafeHandle<Statement> notStatement( new NotStatement(conditionStatement) );

          conditionStatement = notStatement;

       }
     
   }

   StringExpr::skipSeparators(f,l);

   if(f == l)
   {
        if(!handleLineContinuations(stream,
                                    command
                              )
        )
        {
            return false;
        }

        f = command.text_.begin();
        l = command.text_.end();

   }

   string::const_iterator beforeDo = f;

   string doKeyword = parseFirstWord(f,l);

   if(doKeyword != "do")
   {
       messageFiles.writeBytes(2, ErrorMessageWithLocation("while-statement-syntax:  missing 'do' keyword", beforeDo, command) + nl);
       return false;
   }

   // now we have the do keyword, so lets rock and roll


   auto_ptr<StatementList> statementList(new StatementList);

   auto_ptr<StatementListStatement>  statementListStatement;

   {
       // keep lines together for exception safety

       statementListStatement.reset( new StatementListStatement( statementList.get() ) );
       statementList.release();
   }

   static string doneKeyword("done");

   EndOfStatementListDetector endDetector(doneKeyword);


   string::const_iterator saveF;

   bool doneKeywordFound = false;

   auto_ptr<RedirType::REDIRECTION_LIST> redirections;

   while(!stream.eof())
   {
      string::const_iterator saveF = f;
      ParseStatementReturnCode rc = parseStatement(f, l, stream, *statementListStatement, messageFiles, command, endDetector, parsingContext);

      if (rc == PSRC_BAD)
      {
          messageFiles.writeBytes(2, ErrorMessageWithLocation("while-statement-syntax:  missing 'done' keyword",
              f, command) + nl);
         return false;  // io error or syntax error
      }
      else
      if(rc != PSRC_GOOD)
      {
         string::const_iterator doneLoc = f;

         string terminatingKeyword = parseFirstWord(f,l);  // eat the "done" keyword.

         if(terminatingKeyword ==  doneKeyword)
         {
            doneKeywordFound = true;
         }

         StringExpr::skipSeparators(f,l);

         if (f != l && *f == '#')
         {
             while (f != l && *f != '\n')
                ++f;
         }

         pair<string,RedirType::REDIRECTION_LIST*> rd = RedirType::parseOnlyRedirections(f,l);

         if(rd.second)
             redirections.reset(rd.second); // a redirection was actually returned
         else
         {
            // something that was not a redirection was encounter, and it was not and end of statement delimiter or end of block delimiter

            messageFiles.writeBytes(2, ErrorMessageWithLocation("while-statement-syntax:  only i/o redirections are allowed after 'done' keyword",
                                                                doneLoc, command) + nl);
            return false;
         }
         
         StringExpr::skipSeparators(f,l);

         while(f != l)
         {
             saveF = f;

             auto_ptr<StringExpression> token( StringExpr::parse(f, l) );

             if(!token.get())
                break;
             
             if(token->isSpecial())  // propably the end of token
             {
                 string text = token->stringify();

                 if(text == ";;")
                 {
                    f = saveF;
                    break;
                 }

                 if(text[0] == ';')
                 {
                    break;
                 }

                 f = saveF;
                 break;
             }
             
             StringExpr::skipSeparators(f,l);
         }

         break;  // hit end of statement list
      }

      StringExpr::skipSeparators(f,l);

      if(f == l)
      {
          if (!handleLineContinuations(stream, command))
          {
              if (stream.error())
                  return false;
              break;
          }

          f = command.text_.begin();
          l = command.text_.end();
      }

   }

  // if we get here then we have successfully parsed the syntax and the conditionStatement holds the condition to execute
  // and the statementListStatement holds the body.

  if(!doneKeywordFound)
  {
       messageFiles.writeBytes(2, parsingFilename + colon + lexical_cast<string>(parsingStartLine) + colon +
                                  string("while-statement-syntax:  missing 'done' keyword") + nl);
       return false;
  }

  auto_ptr<BourneWhileStatement> whileStatement;

  {
     // keep together for exception saftey

     whileStatement.reset(  new BourneWhileStatement(conditionStatement, ThreadSafeHandle<Statement>(statementListStatement.get()) ) );
     statementListStatement.release();
  }

  whileStatement->setRedirections(redirections.release());


   {
      // for exception saftey, do not separate the following lines
      
       output.statements_->push_back(ThreadSafeHandle<Statement>(whileStatement.get()));
       whileStatement.release();
   }


   

  return true;

}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseForStatement( string::const_iterator           &f, 
                               string::const_iterator           &l,
                               ScriptLineStream                 &stream,
                               StatementListStatement           &output,
                               ScriptThread::ManagedFileHandles &messageFiles,  
                               ScriptInputLine                  &command,   // over which f and l iterate
                               ScriptThread::ThreadContext      *parsingContext
                             )
    ///  Parse either for or select statements from the stream defined by f, l, and stream
{
  string previousPrompt = stream.getPrompt();


   AutoPoppingPushPrompt  prompt(stream, previousPrompt + "-for> " );  // will auto-restore on return

   // for statements have the following syntax:
   //
   //   for f in word ... ; do   statementList ; done  1>&2 2>nul 0<nul
   //
   // note that ; and \n are synonymous in this example.
   
   string startFileName = stream.filename();
   int    startLineNumber = stream.lineNumber();  // for CXXTLS_DEBUG, below

   bool isArithmetic = false; // set to true only for "for (( i=1; i < 100; ++i ))" style statements

   auto_ptr<RedirType::REDIRECTION_LIST> redirections;


   StringExpr::skipSeparators(f,l);

   string::const_iterator statementStart = f;

   string firstWord = parseFirstWord(f,l);

   bool isSelect = false;


   if(firstWord == "select")
   {
      stream.changePrompt(previousPrompt + "-select> ");

      isSelect = true;
   }
   else
   if(firstWord != "for")
   {
      messageFiles.writeBytes(2, ErrorMessageWithLocation("missing 'for' keyword", statementStart, command) + nl);
      return false;
   }

   StringExpr::skipSeparators(f,l);

   if(f==l)
   {
      messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error:  missing loop var name", statementStart, command) + nl);
      return false;
   }

   statementStart = f;

   typedef std::list<cxxtls::StringExpression*> PATTERN_LIST;

   auto_ptr<PATTERN_LIST> patterns(new PATTERN_LIST);

   bool semicolon = false;  // for loop iteration list ends in semicolon

   char c = *f;

   if (f != l && c == '(')
   {

       // we probably have syntax like this:   for (( int i; i < 10 ; ++i )) ...

       string::const_iterator np = f;  
       ++np;

       if (np != l && *np == '(')
       {
           f = np+1; 

           isArithmetic = true;

           string::const_iterator scan = f;

           // scan for the closing )) that delimits the for statement loop criteria

           while (scan != l)
           {  // search for ))

               if (*scan == ')')
               {
                   string::const_iterator next = scan;
                   ++next;

                   if (next == l)
                       break;
                   else
                   if (*next == ')')
                   {
                       // found the end of the loop criteria, parse the 3 expressions in string that runs
                       // from f to scan

                       {
                           string::const_iterator exp1Start = f;   
                           string::const_iterator exp1End   = exp1Start;

                           while (exp1End != scan && *exp1End != ';')
                              ++exp1End;

                           if (exp1End == scan)
                           {
                               messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error:  missing first ; in for loop range definition", statementStart, command) + nl);
                               return false;
                           }

                           string::const_iterator exp2Start = exp1End + 1;
                           string::const_iterator exp2End   =  exp2Start;

                           while (exp2End != scan && *exp2End != ';')
                               ++exp2End;
                           
                           if (exp2End == scan)
                           {
                               messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error:  missing second ; in for loop range definition", statementStart, command) + nl);
                               return false;
                           }

                           string::const_iterator exp3Start = exp2End+1;
                           string::const_iterator exp3End = scan;

                           string loopVarText;   loopVarText.assign(exp1Start, exp1End);
                           string loopCondText;  loopCondText.assign(exp2Start, exp2End);
                           string loopIncText;   loopIncText.assign(exp3Start, exp3End);


                           {
                               // keep this whole block together for exception saftey
                               auto_ptr<cxxtls::StringExpression> loopVarDefinition (new cxxtls::StringExpression());
                               auto_ptr<cxxtls::StringExpression> loopCondition     (new cxxtls::StringExpression());
                               auto_ptr<cxxtls::StringExpression> loopIncrement     (new cxxtls::StringExpression());
                               //
                               //
                               loopVarDefinition->adoptNode  (new cxxtls::StringExpr::LiteralNode(loopVarText));
                               loopCondition->adoptNode      (new cxxtls::StringExpr::LiteralNode(loopCondText));
                               loopIncrement->adoptNode      (new cxxtls::StringExpr::LiteralNode(loopIncText));
                               //
                               //
                               patterns->push_back(loopVarDefinition.get());
                               patterns->push_back(loopCondition.get());
                               patterns->push_back(loopIncrement.get());
                               //
                               loopVarDefinition.release();
                               loopCondition.release();
                               loopIncrement.release();
                           }



                       }

                       // now validate the syntax on the rest of the line

                       f = next;
                       ++f;
                       StringExpr::skipSeparators(f, l);

                       if (f == l)
                          break;  // end of line as expected
                       else
                       if (*f == ';')
                       {
                          semicolon = true;
                          ++f;
                          break;  // statement separator, treat as end of line
                       }
                       else
                       {
                           messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error:  trash after )) in for-statement loop criteria", statementStart, command) + nl);
                           return false;
                       }
                   }
                   else
                   {
                      ++scan;
                   }
               }
               else
                   ++scan;
           } // search for ))

           // this code is current broken because I have not implemented the parser for the arithmetic expressions here.

       }
       else
       {
           messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error:  invalid loop var syntax", statementStart, command) + nl);
           return false;
       }
   }
   else
   if(!isalpha(c) && c != '_')
   {
      messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error:  invalid loop var name", statementStart, command) + nl);
      return false;
   }

   string loopVariableName;
   string::const_iterator saveF = f;


   if (!isArithmetic)
   {
       // parse non-arithmetic syntax

       loopVariableName  = parseFirstWord(f,l);

       saveF = f;

       string inKeyword = parseFirstWord(f,l);

        if(inKeyword != "in")
        {
            messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error: missing 'in' clause", saveF, command) + nl);
            return false;
        }

        
       while(f != l)
       {
          StringExpr::skipSeparators(f,l);

          if(f == l)
          {
             break;
          }

          saveF = f;

          auto_ptr<cxxtls::StringExpression> word( cxxtls::StringExpr::parse(f,l) );

          if(!word.get())
          {
              messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error:  invalid words list",
                                                                  saveF, command) + nl );
              return false;
          }

          string wordValue = word->stringify();

          if(wordValue.empty())
          {
              messageFiles.writeBytes(2, ErrorMessageWithLocation("synax-error: invalid string expression in words list", saveF, command) + nl );
              return false;
          }

          if(wordValue[0] == ';')
          {
              semicolon = true;
              break;
          }

          {
              // for exception saftey, do not separate the following lines 
              patterns->push_back( word.get() );
              word.release();
          }

       }

   }
   

   // if we get here, patterns is populated (or empty) and we have good syntax

   if(!semicolon)
   {
      // we need to get the next line of input from the stream.

      if(!handleLineContinuations(stream, command))
          return false;  // do not ignore i/o errors and empty lines

      f = command.text_.begin();
      l = command.text_.end();

   }

   saveF = f;

   string doKeyword = parseFirstWord(f,l);

   if(doKeyword != "do")
   {
       messageFiles.writeBytes(2, ErrorMessageWithLocation("syntax-error:  expected 'do' keyword", saveF, command) + nl );
        
       return false;
   }

   stream.changePrompt(previousPrompt + "-do> ");

   auto_ptr<StatementList> statementList(new StatementList);

   auto_ptr<StatementListStatement>  statementListStatement;

   {
       // keep lines together for exception safety

       statementListStatement.reset( new StatementListStatement( statementList.get() ) );
       statementList.release();
   }

   static string doneKeyword("done");

   EndOfStatementListDetector endDetector(doneKeyword);


   while(!stream.eof())
   {
      ParseStatementReturnCode rc = parseStatement(f, l, stream, *statementListStatement, messageFiles, command, endDetector, parsingContext);

      if(rc == PSRC_BAD)
         return false;  // io error or syntax error
      else
      if(rc != PSRC_GOOD)
      {
         parseFirstWord(f,l);  // eat the "done" keyword.

         StringExpr::skipSeparators(f,l);

         if (f != l && *f == '#')
         {
             while (f != l && *f != '\n')
                ++f;
         }

         // now parse the trailing redirections

         pair<string, RedirType::REDIRECTION_LIST*> rd = RedirType::
                                                        parseOnlyRedirections(f,
                                                                              l
                                                                             );
         if(rd.second)
         {
            redirections.reset(rd.second);
         }
         else
         {
              messageFiles.writeBytes(2, 
              string("Error:  i/o synax error after 'done' keyword:  ") + rd.first + nl);
              return false;
            
         }

         StringExpr::skipSeparators(f,l);

         while(f != l)
         {
             saveF = f;

             auto_ptr<StringExpression> token( StringExpr::parse(f, l) );

             if(!token.get())
                break;
             
             if(token->isSpecial())  // propably the end of token
             {
                 string text = token->stringify();

                 if(text == ";;")
                 {
                    f = saveF;
                    break;
                 }

                 if(text[0] == ';')
                 {
                    break;
                 }

                 f = saveF;
                 break;
             }
             
             StringExpr::skipSeparators(f,l);
         }

         break;  // hit end of statement list
      }

      StringExpr::skipSeparators(f,l);

      if(f == l)
      {
          if (!handleLineContinuations(stream, command))
          {
              if (stream.error())
                  return false;
              break;
          }

          f = command.text_.begin();
          l = command.text_.end();
      }

   }


   auto_ptr<BourneForStatement> bourneForStatement;


   {
      // for exception saftey, do not separate the following lines

      bourneForStatement.reset(   new BourneForStatement( loopVariableName, *patterns, statementListStatement.get(), isSelect, isArithmetic ) );

      statementListStatement.release();

   }

   bourneForStatement->setRedirections(redirections.release());


   CXXTLS_DEBUG(startFileName << ":" << startLineNumber << ":  creating FOR statement");

   {
      // for exception saftey, do not separate the following lines
      
       output.statements_->push_back(ThreadSafeHandle<Statement>(bourneForStatement.get()));
       bourneForStatement.release();
   }

   return true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


static bool parseAliasStatement( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                            )
{

   // f and l iterate over command

   string::const_iterator startOfLine = f;

   string aliasKeyword = parseFirstWord(f,l);

   if(aliasKeyword != "alias")
   {
       messageFiles.writeBytes(2, ErrorMessageWithLocation("alias-syntax:  missing alias keyword", startOfLine, command) + nl );
       return false;
   }

   StringExpr::ExecutionContext::AliasMap  aliases;


   StringExpr::skipSeparators(f, l);

   if(f == l)
   {
      // no options provided -- list the existing aliases

      messageFiles.writeBytes(1, string("Existing Aliases:\r\n"));

      typedef StringExpr::ExecutionContext::AliasMap AliasMap;

      CXXTLS_FOREACH(AliasMap::value_type const &cur, parsingContext->variables().aliases())
      {
         messageFiles.writeBytes(1, space + space + cur.first + equalSign);
         
         CXXTLS_FOREACH(string const &impl, cur.second)
         {
             messageFiles.writeBytes(1, impl + space);
         }

         messageFiles.writeBytes(1, nl);

      }
      return true;

   }


   while(f != l)
   {
       startOfLine = f;
       string aliasName = parseFirstWord(f,l);
       
       if (aliasName.empty())
       {
           messageFiles.writeBytes(2, ErrorMessageWithLocation("alias-syntax:  missing name", startOfLine, command) + nl);
           return false;
       }

       startOfLine = f;
       StringExpr::skipSeparators(f,l);
       
       if(f == l || *f != '=')
       {
            messageFiles.writeBytes(2, ErrorMessageWithLocation("alias-syntax:  missing '='", startOfLine, command) + nl );
            return false;
       }
       
       ++f;  // skip =

       StringExpr::skipSeparators(f,l);

       if(f == l)
       {
           messageFiles.writeBytes(2, ErrorMessageWithLocation("alias-syntax:  value to complex", startOfLine, command) + nl);
           return false;
       }


       startOfLine = f;

       auto_ptr<StringExpression> value(StringExpr::parse(f,l));
       
       string aliasValue;

       if(!value->isSimple(aliasValue))
       {
           messageFiles.writeBytes(2, ErrorMessageWithLocation("alias-syntax:  value to complex", startOfLine, command) + nl);
           return false;
       }


       StrTool::stringlist_t parsedValue;

       StrTool::parse_words(aliasValue, &parsedValue, 100000, StrTool::Is_Space());

       StringExpr::ExecutionContext::AliasMap::iterator where = aliases.find(aliasName);

       if(where == aliases.end())
       {
          aliases[aliasName] = StringExpr::ExecutionContext::ArgsVector();
          where = aliases.find(aliasName);
       }
       else
          where->second.clear();  // erase existing alias contents

       CXXTLS_FOREACH(string const &cur, parsedValue)
       {
           where->second.push_back(cur);
       }
       StringExpr::skipSeparators(f, l);
   }

   // at this point, we have the alias map properly parsed

   if(parsingContext == 0)
   {
      messageFiles.writeBytes(2, string("Error:  alias command used without a parsing context?   "));
   }
   else
   {
       CXXTLS_FOREACH(StringExpr::ExecutionContext::AliasMap::value_type const &cur, aliases)
       {
           parsingContext->variables().addAlias(cur.first, cur.second);
       }
   }



   return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



static bool parseUnAliasStatement( string::const_iterator           &f, 
                                   string::const_iterator           &l,
                                   ScriptLineStream                 &stream,
                                   StatementListStatement           &output,
                                   ScriptThread::ManagedFileHandles &messageFiles,
                                   ScriptInputLine                  &command,   // over which f and l iterate
                                   ScriptThread::ThreadContext      *parsingContext
                            )
{
   // f and l iterate over command

   string::const_iterator startOfLine = f;

   string aliasKeyword = parseFirstWord(f,l);

   if(aliasKeyword != "unalias")
   {
       messageFiles.writeBytes(2, ErrorMessageWithLocation("unalias-syntax:  'unalias' keyword missing", startOfLine, command) + nl );
       return false;
   }

   StringExpr::skipSeparators(f, l);

   if(f == l)
   {
      // no options provided -- list the existing aliases

      messageFiles.writeBytes(1, string("Existing Aliases:\r\n"));

      typedef StringExpr::ExecutionContext::AliasMap AliasMap;

      CXXTLS_FOREACH(AliasMap::value_type const &cur, parsingContext->variables().aliases())
      {
         messageFiles.writeBytes(1, space + space + cur.first + equalSign);
         
         CXXTLS_FOREACH(string const &impl, cur.second)
         {
             messageFiles.writeBytes(1, impl + space);
         }

         messageFiles.writeBytes(1, nl);

      }
      return true;

   }


   while(f != l)
   {
       startOfLine = f;
       string aliasName = parseFirstWord(f,l);
       
       if (aliasName.empty())
       {
           messageFiles.writeBytes(2, ErrorMessageWithLocation("unalias-syntax:  missing alias name", startOfLine, command) + nl );
           return false;
       }

       parsingContext->variables().removeAlias(aliasName);

       StringExpr::skipSeparators(f,l);

   }



   return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


static bool parseNotStatement  ( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               );


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


static bool parseIfStatement   ( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               );

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


static bool parseCaseStatement ( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               );


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


static bool parseBlockStatement( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               );


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool isFuncName(string const &s)
{
   CXXTLS_FOREACH(char c, s)
   {
      if(   !isalnum(c)
         && c == '_'
        )
      {
         return false;
      }
   }

   return true;
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

 ParseStatementReturnCode parseStatementBody  (
                                               string::const_iterator           &f,
                                               string::const_iterator           &l,
                                               ScriptLineStream                 &stream,
                                               StatementListStatement           &output,
                                               ScriptThread::ManagedFileHandles &messageFiles,
                                               ScriptInputLine                  &command,
                                               EndOfStatementListDetector const &endDetector,
                                               ScriptThread::ThreadContext      *parsingContext
                                              )
    ///  parse lines from stream that comprise a single bash statement
{

    StringExpr::skipSeparators(f,l);

    string::const_iterator commandStart = f;


    // this code is definitely going to stop on end of file.  It may stop
    // on a specified keyword or character as the first thing found on a command line
    // based on whether or not endDetector flags tells us to do so.

    bool endOfStatementList = false;


        if(endDetector.endsWithCharacter_ && *f == endDetector.endingCharacter_)
        {
           f = commandStart;
           return PSRC_EOL;   // found ), }, or some other end of list indicator.
        }

        string::const_iterator beforeFirstWord = f;

        string firstWord = parseFirstWord(f, l);

        if (!firstWord.empty())
        {
            // check for do, done, else, fi, and later for other special tokens

            char c = firstWord[0];

            if (   (c == 'd' && (firstWord == "do" || firstWord == "done"))
                || (c == 'e' && firstWord == "else")
                || (c == 'f' && firstWord == "fi")
                || (c == 't' && firstWord == "then")
               )
            {
                if (f == l || *f != '=')
                {
                    // we have an "ending" keyword that is not contiguous with a trailing '='
                    f = beforeFirstWord;
                    return PSRC_EOL;
                }
            }
        
        }




        if (f != l && *f == '!')
        {
            firstWord = "!";
        }
        else
        if(firstWord.empty())
        {
            if (f != l && *f == ';')
            {
                return PSRC_GOOD;
            }
           else
           if(f != l && *f != '#')
           {
              messageFiles.writeBytes(2, ErrorMessageWithLocation("Bad character at line start", commandStart, command) + nl);


              return PSRC_BAD;
           }

           if (f != l && *f == '#')
               f = l; // eat the comment

           return PSRC_GOOD; // just ignore empty lines and lines begining with a comment
        }
        else
        if(   endDetector.endsWithKeyword_
           && endDetector.isEndingKeyword(firstWord) 
           && (
                  f == l
                  || *f != '='
              )
          )
        {
           // we get here when a keyword is found but not if the keyword is contiguous with a trailing '='

           f = commandStart;
           return PSRC_EOL;   // found done, else, fi, or some other end of list indicator.
        }
    
        // if we get here, the current line has at least one word one it -- including the possibility
        // that the first word contains ( or {
    
        char nc = 0;
    
        if (f != l) 
           nc = *f;
    
        char fc = firstWord[0];
    
        f = commandStart;  // set up to restart the parsing process

        if(fc == '#')
        {
           f = l;        // eat the rest of the command line
           return PSRC_GOOD;  // just ignore comments
        }
    
        bool good = false;

        bool isFuncDecl = false;   // will be true if we have:   "firstWord ()"

        // check to see this syntax:    "firstWord()", tightly together

        size_t openParenOffset = firstWord.size();
        size_t closeParenOffset= firstWord.size();

        closeParenOffset = firstWord.find_first_of(')');

        if(!firstWord.empty() && closeParenOffset < firstWord.size())
        {
           openParenOffset = firstWord.find_first_of('(');

           if(closeParenOffset == openParenOffset + 1)
           {
              char c = firstWord[0];

              if( isalnum(c) || c == '_' ) 
                 isFuncDecl = true;
           }
        }

        if(fc != '!' && !isFuncDecl && !firstWord.empty() && isFuncName(firstWord))
        {
            // now see if we have "funcname ()"

            string::const_iterator  g = f;
            StringExpr::skipSeparators(g, l);

           parseFirstWord(g,l);  // reparse the first word.  sorry, that's the way this works

           StringExpr::skipSeparators(g, l);

           if(g != l && *g == '(')
           {
              string::const_iterator next = g;

              ++next;

              if(next != l && *next == ')')
              {
                 isFuncDecl = true;
              }
           }
           
        }

        if(isFuncDecl)
        {
           good = parseFunction(f, l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(fc == '{' || fc == '(')
        {
           if(firstWord == "((")
           {

               good =  parseArithmeticStatement(f,l, stream, output, messageFiles, command);

               if(good && f != l)
               {

                   StringExpr::skipSeparators(f,l);

                   string::const_iterator postParens = f;


                   if(     f != l
                      &&  *f != '&'
                      &&  *f != ';'
                      &&  *f != '#'
                      &&  *f != '|'
                     )
                   {
                      good = false;  // only end of statement or comment can come after ))

                      messageFiles.writeBytes(2, ErrorMessageWithLocation("))-syntax:  trailing tokens not allowed", postParens, command) + nl);

                   }
               }

           }
           else
               good =  parseBlockStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(nc  == '=' && !firstWord.empty())
        {
           // we either have an assignment statement or we have a set of variable superseeds
           // before command line statement.
    
           good = parseCommandLine(f,l, stream, output, messageFiles, command, parsingContext);
    
        }
        else
        if(fc== 'i' && firstWord == "if")
        {
           good =  parseIfStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(fc == 'w' && firstWord == "while")
        {
           good =  parseWhileStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(fc == 'u' && firstWord == "until")
        {
           // parseWhileStatement also handles "until" statements
           good =  parseWhileStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(fc == 'f' && firstWord == "for")
        {
           good =  parseForStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(fc == 's' && firstWord == "select")
        {
           good =  parseForStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(fc == 'a' && firstWord == "alias")
        {
           good =  parseAliasStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(fc == 'u' && firstWord == "unalias")
        {
           good =  parseUnAliasStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if(fc == 'f' && firstWord == "function")
        {
           good =  parseFunction(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if( fc == 'c' && firstWord == "case" )
        {
          good = parseCaseStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if( fc == '!' && firstWord == "!")
        {
          good = parseNotStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        if( fc == 'f' && firstWord == "fc")
        {
          good = parseFcStatement(f,l, stream, output, messageFiles, command, parsingContext);
        }
        else
        {
            good =  parseCommandLine(f,l, stream, output, messageFiles, command, parsingContext);
        }

        if(!good)
           return PSRC_BAD;




    return PSRC_GOOD;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

 ParseStatementReturnCode parsePipeStatement(
                                               string::const_iterator           &f,
                                               string::const_iterator           &l,
                                               ScriptLineStream                 &stream,
                                               StatementListStatement           &output,
                                               ScriptThread::ManagedFileHandles &messageFiles,
                                               ScriptInputLine                  &command,
                                               EndOfStatementListDetector const &endDetector,
                                               ScriptThread::ThreadContext      *parsingContext
                                              )
    ///  parse all the statemehts from stream that comprise a bash statement that can be part of a pipe.
{
    ParseStatementReturnCode rc=PSRC_GOOD;

   std::list< ThreadSafeHandle<Statement> > pipeChain;


    try
    {
    
        while(f != l)
        {
            skipEmptyStatements(f, l, stream, command);

            if(f == l)
               break; // eof

            size_t beginningSize = output.statements_->size();
    
            rc = parseStatementBody(f,l,stream,output,messageFiles,command,endDetector,parsingContext);
    
            if(rc != PSRC_GOOD)
               break;
    
            size_t endingSize =  output.statements_->size();
    
            size_t addedStatements = endingSize  - beginningSize;
    
            if(addedStatements > 1)
            {
                throw ParseStatementInternalCheckFailedException("parseStatementBody produced too many statements ", addedStatements);
            }
    
            StringExpr::skipSeparators(f,l);
    
            string op; // the current statement's terminator.  Could be end of line or it could be ;, or ;;, or |, or ||, or &, or &&, etc.

            string::const_iterator startOfTerminator = f;
    
            while(f != l && (*f == ';' || *f == '&' || *f == '|'))
               op += *f++;
    
            StringExpr::skipSeparators(f,l);
    
            if(op == "||" || op == "&&" || op == ";;")
            {
                f = startOfTerminator;
                return rc;  // should be PSRC_GOOD
            }

           
            if(op.empty() || op == ";" || op == "&" )
            {
               if(!pipeChain.empty())
               {
                  // if we have a pipe chain at then end of the statement, we must
                  // convert the most recently parsed statement into the end of the
                  // chain and incorporate the chain into pipe statement which replaces
                  // the most recently parsed item

                  {
                  
                      CXXTLS_DEBUG("ParseStatement: adding pipe statement to chain(" << std::hex << (long )(&*output.statements_->back()) << ")");  


                      {
                          // keep all this together
                          pipeChain.push_back(output.statements_->back());
                          output.statements_->pop_back();
                      }
                  }

                  // the most recent command is now pushed into the pipe chain

                  output.statements_->push_back( ThreadSafeHandle<Statement>(new BournePipeStatement(pipeChain)) );

                  if(!pipeChain.empty())
                  {
                     // huh?
                    throw ParseStatementInternalCheckFailedException("BournePipeStatement constructor did not consume pipeChain contents ","");
                  }

               }

               if (op == "&")
               {
                   output.statements_->back()->setDetached(true);
               }

               break;
            }

            if(op == "|")
            {
                {
                    // keep all this together
                    pipeChain.push_back(output.statements_->back());
                    output.statements_->pop_back();

                    // skip blanks and comments after the | symbol
                    StringExpr::skipSeparators(f, l);

                    if (f != l && *f == '#')
                    {
                        while (f != l)
                           ++f;
                    }

                    if (f == l)
                    {
                    
                        if (!handleLineContinuations(stream, command))
                        {
                            f = l;
                        }
                        else
                        {

                            f = command.text_.begin();
                            l = command.text_.end();
                        }
                    }

                }
               
            }

            if (op == ";;")
            {
                f = startOfTerminator;
                break;
            }
        }
    }
    catch(...)
    {
       throw;   // anything weird happens, delete the pipe contents, if any
    }

    return rc;

} // parsePipeStatement()

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

 ParseStatementReturnCode parseAndAndStatement(
                                               string::const_iterator           &f,
                                               string::const_iterator           &l,
                                               ScriptLineStream                 &stream,
                                               StatementListStatement           &output,
                                               ScriptThread::ManagedFileHandles &messageFiles,
                                               ScriptInputLine                  &command,
                                               EndOfStatementListDetector const &endDetector,
                                               ScriptThread::ThreadContext      *parsingContext
                                              )
    ///  parse lines from stream that comprise a bash statement that can be part of a pipe.
{

    size_t beginningSize;
    
    ParseStatementReturnCode rc;
    
    size_t endingSize;
    
    size_t addedStatements;
    
    while(f != l)
    {
        if (*f == ';' && f != l)
        {
            // check for the start of statement being ';;'

            string::const_iterator g = f; ++g;

            if (g != l && *g == ';')
               return PSRC_GOOD; 
        }

        beginningSize = output.statements_->size();
        
        rc = parsePipeStatement(f,l,stream,output,messageFiles,command,endDetector,parsingContext);
        
        if(rc != PSRC_GOOD)
           return rc;
        
        endingSize =  output.statements_->size();
        
        addedStatements = endingSize  - beginningSize;

        if(addedStatements > 1)
        {
            throw ParseStatementInternalCheckFailedException("parsePipeStatement produced too many statements ", addedStatements);
        }
        else
        if(addedStatements < 1)
        {
            // we can get here if the source contains a statement that executes at parse time rather than run time.
            // the alias statement is one of those kind

            continue;
        }

        break; 

    }

    // if we get here, it means that we have exactly 1 statement inserted into output by the parsePipeStatement loop above


    while (f != l)
    {
    
        string::const_iterator startOfTerminator = f;

        string op;
    
        while(f != l && (*f == ';' || *f == '&' || *f == '|'))
           op += *f++;
    
        StringExpr::skipSeparators(f,l);
    
        if(op != "&&" )
        {
            f = startOfTerminator;
            return PSRC_GOOD;  // not a && 
        }

        // if we get here, we are dealing with && or ||

        // the most statement pushed onto the output will become the left hand side of our conditional statement
        // and we will parse the right hand side then combine them together, then push the result back onto the
        // output so that it pair can then serve as the left hand side of another && or || statement, if needed.

        ThreadSafeHandle<Statement> lhs( output.statements_->back() );
        output.statements_->pop_back();


        beginningSize = output.statements_->size();
        
        rc = parsePipeStatement(f,l,stream,output,messageFiles,command,endDetector,parsingContext);
        
        if(rc != PSRC_GOOD)
           return rc;
        
        endingSize =  output.statements_->size();
        
        addedStatements = endingSize  - beginningSize;
        
        if(addedStatements > 1)
        {
            throw ParseStatementInternalCheckFailedException("parsePipeStatement produced too many statements ", addedStatements);
        }
        else
        if(addedStatements < 1)
        {
            throw ParseStatementInternalCheckFailedException("parsePipeStatement produced no statement while parsing && ", addedStatements);
        }
    
        // keep together for safety
        ThreadSafeHandle<Statement> rhs( output.statements_->back() );
        output.statements_->pop_back();

        auto_ptr<Statement> tmp(new BourneConditionalStatement(lhs, rhs, op[0] == '&') );  // && or || statements based on the operator

        {
           // keep together for exception safety.
           output.statements_->push_back(ThreadSafeHandle<Statement>(tmp.get()));
           tmp.release();
        }

        // ready for round two..
    
        
    }

    return PSRC_GOOD;

} // parseAndAndStatement


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

 ParseStatementReturnCode parseStatement(
                                               string::const_iterator           &f,
                                               string::const_iterator           &l,
                                               ScriptLineStream                 &stream,
                                               StatementListStatement           &output,
                                               ScriptThread::ManagedFileHandles &messageFiles,
                                               ScriptInputLine                  &command,
                                               EndOfStatementListDetector const &endDetector,
                                               ScriptThread::ThreadContext      *parsingContext
                                              )
    ///  parse lines from stream that comprise a bash statement.
    ///
    ///  also known as "parse or or statement", from a syntax point of view
{

    CXXTLS_DEBUG("parseStatement:  parsingContext = " << (void*)(parsingContext) );

    if(parsingContext)
    {
        CXXTLS_DEBUG("parseStatement:  executionContext = " << (void*)(&parsingContext->variables()) );
        CXXTLS_DEBUG("parseStatement:  executionVariables = " << (void*)(&parsingContext->variables().variables()) );
    }

    size_t beginningSize;
    
    ParseStatementReturnCode rc;
    
    size_t endingSize;
    
    size_t addedStatements = 0;
    
    while(f != l)
    {
        if (*f == ';' && f != l)
        {
            // check for the start of statement being ';;'

            string::const_iterator g = f; ++g;

            if (g != l && *g == ';')
                return PSRC_GOOD;
        }

        beginningSize = output.statements_->size();
        
        rc = parseAndAndStatement(f,l,stream,output,messageFiles,command,endDetector,parsingContext);
        
        if(rc != PSRC_GOOD)
           return rc;
        
        endingSize =  output.statements_->size();
        
        addedStatements = endingSize  - beginningSize;

        if(addedStatements != 1)
        {
            continue;
        }

        break; 

    }

    // if we get here, it means that we have exactly 1 statement inserted into output by the parsePipeStatement loop above


    while (f != l)
    {
    
        string::const_iterator startOfTerminator = f;

        string op;
    
        while(f != l && (*f == ';' || *f == '&' || *f == '|'))
           op += *f++;
    
        StringExpr::skipSeparators(f,l);
    
        if(op != "||" )
        {
            f = startOfTerminator;
            return PSRC_GOOD;  // not a && 
        }

        // if we get here, we are dealing with && or ||

        // the most statement pushed onto the output will become the left hand side of our conditional statement
        // and we will parse the right hand side then combine them together, then push the result back onto the
        // output so that it pair can then serve as the left hand side of another && or || statement, if needed.

        ThreadSafeHandle<Statement> lhs( output.statements_->back() );
        output.statements_->pop_back();


        beginningSize = output.statements_->size();
        
        rc = parseAndAndStatement(f,l,stream,output,messageFiles,command,endDetector,parsingContext);
        
        if(rc != PSRC_GOOD)
           return rc;
        
        endingSize =  output.statements_->size();
        
        addedStatements = endingSize  - beginningSize;
        
        if(addedStatements != 1)
        {
            throw ParseStatementInternalCheckFailedException("parseAndandStatement produced the wrong number of statements ", addedStatements);
        }
    
        // keep together for safety
        ThreadSafeHandle<Statement> rhs( output.statements_->back() );
        output.statements_->pop_back();

        auto_ptr<Statement> tmp(new BourneConditionalStatement(lhs, rhs, op[0] == '&') );  // && or || statements based on the operator

        {
           // keep together for exception safety.
           output.statements_->push_back(ThreadSafeHandle<Statement>(tmp.get()));
           tmp.release();
        }

        // ready for round two..
    
        
    }

    if(addedStatements)
    {
       // handle inline input redirections by converting the most recently parsed statement into a mini-pipe
       //
       //   We are parsing something like this:
       //
       //    statement <<*eof
       //         line 1
       //         line 2
       //    *eof
       //
       // Note that the 'statement' is parsed above, and at this point the code we are parsing the line1 -> *eof


       if(output.statements_->empty())
       {
           cerr << __FILE__ << __LINE__ << ":  Brash Fatal Internal Error\n";
           exit(1);
       }

       ThreadSafeHandle<Statement> finalStatement = output.statements_->back();

       cxxtls::StringExpression const* inlineRedir = findInlineInputRedirection(&*finalStatement);

       if(inlineRedir)
       {
           string tag;

           tag = inlineRedir->stringify();

           if(tag.size() > 2)
           {

               AutoPoppingPushPrompt  promptTmp(stream, "data> ");


               tag = tag.substr(2);

               // now read lines from the command stream until we find tag or end of file

               size_t chars = tag.size();

               list<string>   input;

               if (!handleLineContinuations(stream, command))
               {
                  f = l;
               }
               else
               {

                   f = command.text_.begin();
                   l = command.text_.end();
               }

               while(f != l)
               {
                   string curline;

                   // read a line

                   while(f != l)
                   {
                       if (*f == '\r')
                       {
                          ++f;
                          continue;
                       }

                       if (*f == '\n')
                       {
                          ++f;
                          break;
                       }

                       curline += *f++;
                   }


                   if(   curline.size() >= tag.size() 
                      && 0 == curline.compare(0, sizeof(tag), tag)
                     )
                      break;  // found end of input tag

                   input.push_back(curline);

                   string filename = stream.filename();
                   size_t lineno   = stream.lineNumber();

                   if(f == l)
                   {
                       // fetch the next line if needed

                       if(!handleLineContinuations(stream, command))
                       {
                           break;
                       }
                   
                       f = command.text_.begin();
                       l = command.text_.end();
                   }

                   if(f == l)
                   {
                       messageFiles.writeBytes(2, filename + lexical_cast<string>(lineno) + string(": Error out of text searching for tag '")
                                                  + tag + string("'\r\n")
                                              );
                       return PSRC_BAD;
                   }
               }
               // now convert the input data into a command which pipe's it to the current statement

               std::list< ThreadSafeHandle<Statement> > pipeChain;

               ThreadSafeHandle<Statement>  writer( new BrashWriteExpressions(input) );

               pipeChain.push_back(writer);

               {  // keep together for exception saftey

                   pipeChain.push_back(output.statements_->back());
                   output.statements_->pop_back();
               }

               output.statements_->push_back( ThreadSafeHandle<Statement>(new BournePipeStatement(pipeChain)) );

           }
       }


    }



    return PSRC_GOOD;

} // parseAndAndStatement

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool parseBrash(ScriptLineStream                 &stream,
                BrashStatementList               &output,
                ScriptThread::ManagedFileHandles &messageFiles,
                ScriptThread::ThreadContext      *parsingContext,
                bool                              wholeFileAsLine   // defaults to false
               )
{
try
{
     AutoPoppingPushPrompt  prompt(stream, stream.getPrompt() + ">" );  // will auto-restore on return

     auto_ptr<StatementList>          statementList(new StatementList);

     auto_ptr<StatementListStatement>  statementListStatement;

     {
         // keep lines together for exception safety

         statementListStatement.reset( new StatementListStatement( statementList.get() ) );
         statementList.release();
     }


     ScriptInputLine command;     // individual parses may refil as the while loop below progresses.
 
     if(!handleLineContinuations(stream, command, wholeFileAsLine))
        return !stream.error();  // do not ignore i/o errors
 
     string::const_iterator f = command.text_.begin(),
                            l = command.text_.end();


     EndOfStatementListDetector endDetector;   // default constructor means "end of file


     while(!stream.eof() || !command.text_.empty())
     {
         ParseStatementReturnCode rc = parseStatement(f, l, stream, *statementListStatement, messageFiles, command, endDetector, parsingContext);

        if(rc != PSRC_GOOD)
           return false;

        StringExpr::skipSeparators(f,l);

        if(f == l)
        {
            if (!handleLineContinuations(stream, command))
            {
                if (stream.error())
                    return false;
                break;
            }

            f = command.text_.begin();
            l = command.text_.end();
        }

     }

     {
        // keep the following lines together for exception safety
        output.statements_->push_back(ThreadSafeHandle<Statement>(statementListStatement.get()));  
        statementListStatement.release();
     }


     return true;

} catch(std::exception const &e)
{

    messageFiles.writeBytes(2, string("Unhandled exception  ") + string(e.what()) + comma + string("while parsing ") + stream.filename() + colon + lexical_cast<string>(stream.lineNumber()) + nl );

    return false;
}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseIfStatement   ( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               )
{
  string previousPrompt = stream.getPrompt();


   AutoPoppingPushPrompt  prompt(stream, previousPrompt + "-if> " );  // will auto-restore on return

   // if statements have the following syntax:
   //
   //   if statement ; then   statementList ; [ else statementList ; ]  fi  1>&2 2>nul 0<nul
   //
   // note that ; and \n are synonymous in this example.

   string parsingFilename = command.filename_;
   int    parsingStartLine= command.lineNumber_;

   string::const_iterator beforeIf = f;

   string firstWord = parseFirstWord(f,l);

   if(firstWord != "if")
   {
      messageFiles.writeBytes(2, ErrorMessageWithLocation("if-statement:  missing 'if' keyword", beforeIf, command) + nl);
      return false;
   }

   StringExpr::skipSeparators(f,l);

   if(f==l)
   {
      messageFiles.writeBytes(2, ErrorMessageWithLocation("if-statement:  missing condition", beforeIf, command) + nl);
      return false;
   }

   ThreadSafeHandle<Statement> conditionStatement;

   {
       // parse the condition statement

       StatementListStatement  conditionTmp(new StatementList);   
       
       // we are going to parse into this and then extract the one statement and put it in conditionStatement.
     
       EndOfStatementListDetector endDetector;

       string::const_iterator beforeCondition = f;

       ParseStatementReturnCode rc = parseStatement(f, l, stream, conditionTmp, messageFiles, command, endDetector, parsingContext);

       if(PSRC_BAD == rc)
       {
           return false;  
       }

       // now move the thing just parsed from the condition temporary list into the conditionStatement variable

       {
           // keep together
           conditionStatement = conditionTmp.statements_->statements().back();
           conditionTmp.statements_->pop_back();  // unlikely to throw, but might
       }
     
   }

   StringExpr::skipSeparators(f,l);

   if(f == l)
   {
        if(!handleLineContinuations(stream,
                                    command
                              )
        )
        {
            return false;
        }

        f = command.text_.begin();
        l = command.text_.end();

   }

   string::const_iterator beforeThen = f;

   string doKeyword = parseFirstWord(f,l);

   if(doKeyword != "then")
   {
       messageFiles.writeBytes(2, parsingFilename + colon + lexical_cast<string>(parsingStartLine) + colon + ":  Error missing 'then' keyword\r\n");
       return false;
   }

   // now we have the do keyword, so lets rock and roll


  stream.changePrompt(previousPrompt+"-then> ");


   auto_ptr<StatementList> mainStatementList(new StatementList);

   auto_ptr<StatementListStatement>  mainStatementListStatement;

   {
       // keep lines together for exception safety

       mainStatementListStatement.reset( new StatementListStatement( mainStatementList.get() ) );
       mainStatementList.release();
   }

   auto_ptr<StatementList> elseStatementList(new StatementList);

   auto_ptr<StatementListStatement>  elseStatementListStatement;

   {
       // keep lines together for exception safety

       elseStatementListStatement.reset( new StatementListStatement( elseStatementList.get() ) );
       elseStatementList.release();
   }


   static string fiKeyword("fi");
   static string elseKeyword("else");

   EndOfStatementListDetector endDetector(fiKeyword, elseKeyword);

   string::const_iterator saveF;

   StatementListStatement *statementListStatement = mainStatementListStatement.get();

   bool fiKeywordFound = false;

   auto_ptr<RedirType::REDIRECTION_LIST> redirections;

   while(!stream.eof())
   {
      ParseStatementReturnCode rc = parseStatement(f, l, stream, *statementListStatement, messageFiles, command, endDetector, parsingContext);

      if(rc == PSRC_BAD)
         return false;  // io error or syntax error
      else
      if(rc != PSRC_GOOD)
      {
         string endingKeyword = parseFirstWord(f,l);  // eat the "fi" keyword.

         if(endingKeyword == "else")
         {
             stream.changePrompt(previousPrompt + "-else>  ");
             statementListStatement = elseStatementListStatement.get();
             continue;
         }
         else
         if (endingKeyword == fiKeyword)
         {
             fiKeywordFound = true;
         }

         string redirFilename = command.filename_;
         int    redirStartline= command.lineNumber_;

         StringExpr::skipSeparators(f,l);

         if (f != l && *f == '#')
         {
             while (f != l && *f != '\n')
                ++f;
         }

         pair<string,RedirType::REDIRECTION_LIST*> rd = RedirType::parseOnlyRedirections(f,l);

         if(rd.second)
             redirections.reset(rd.second);
         else
         {
            messageFiles.writeBytes(2, errorKeyword + redirFilename + colon + lexical_cast<string>(redirStartline) + 
                                       colon + string("if-statement:  only i/o redirections allowed after 'fi'") + nl );
            return false;
         }

         StringExpr::skipSeparators(f,l);

         while(f != l)
         {
             saveF = f;

             auto_ptr<StringExpression> token( StringExpr::parse(f, l) );

             if(!token.get())
                break;
             
             if(token->isSpecial())  // propably the end of token
             {
                 string text = token->stringify();

                 if(text == ";;")
                 {
                    f = saveF;
                    break;
                 }

                 if(text[0] == ';')
                 {
                    break;
                 }

                 f = saveF;
                 break;
             }
             
             StringExpr::skipSeparators(f,l);
         }

         break;  // hit end of statement list
      }

      StringExpr::skipSeparators(f,l);

      if(f == l)
      {
          if (!handleLineContinuations(stream, command))
          {
              if (stream.error())
                  return false;
              break;
          }

          f = command.text_.begin();
          l = command.text_.end();
      }

   }

  // if we get here then we have successfully parsed the syntax and the conditionStatement holds the condition to execute
  // and the statementListStatement holds the body.

   if (!fiKeywordFound)
   {
       messageFiles.writeBytes(2, parsingFilename + colon + lexical_cast<string>(parsingStartLine) + colon + 
                                  string("Error:  if statement missing 'fi' keyword\r\n")
                              );

       return false;
   }
  
  auto_ptr<BourneIfStatement> ifStatement;

  {
     // keep together for exception saftey

     ifStatement.reset(  new BourneIfStatement(conditionStatement, 
                                               ThreadSafeHandle<Statement>(mainStatementListStatement.get()), 
                                               ThreadSafeHandle<Statement>(elseStatementListStatement.get())
                                              ) 
                      );
     mainStatementListStatement.release();
     elseStatementListStatement.release();
  }


   ifStatement->setRedirections(redirections.release());

   {
      // for exception saftey, do not separate the following lines
      
       output.statements_->push_back(ThreadSafeHandle<Statement>(ifStatement.get()));
       ifStatement.release();
   }


   

  return true;

}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseBlockStatement( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               )
{
   AutoPoppingPushPrompt  prompt(stream, stream.getPrompt() + ">" );  // will auto-restore on return
   
   
   if(f == l)
   {
       messageFiles.writeBytes(2, string("Error:  Attempting to parse a block statement at the end of file.\r\n"));
       return false;
   }

   string::const_iterator statementStart = f;
   string startFile = command.filename_;
   int    startLine = command.lineNumber_;

   char statementType = *f++;

   if(f == l)
   {
       string startFilename    = command.filename_;
       int    startLinenumber  = command.lineNumber_;

       if(!handleLineContinuations(stream, command))
       {
           messageFiles.writeBytes(2, errorKeyword + startFilename + colon + lexical_cast<string>(startLinenumber) + colon
                                      + string(" end of file in block starement") + nl );
           return false;
       }

       f = command.text_.begin();
       l = command.text_.end();

       if (f == l)
           return false;
   }

   char terminatingCharacter = (statementType == '{' ? '}' : ')');

   EndOfStatementListDetector  endDetector(terminatingCharacter);

   auto_ptr<StatementList> statementList(new StatementList);

   if(statementType == '(')
       statementList->setSubShellFlag(true);


   auto_ptr<StatementListStatement>  statementListStatement;

   {
       // keep lines together for exception safety

       statementListStatement.reset( new StatementListStatement( statementList.get() ) );
       statementList.release();
   }


   auto_ptr<BrashCommandLine::REDIRECTION_LIST>       redirections (new BrashCommandLine::REDIRECTION_LIST);


   while(!stream.eof())
   {  // read statements

      ParseStatementReturnCode rc = parseStatement(f, l, stream, *statementListStatement, messageFiles, command, endDetector, parsingContext);

      if(rc == PSRC_BAD)
         return false;  // io error or syntax error
      else
      if(rc == PSRC_EOL)
      {
         string::const_iterator endingLineStart =f ;
         
         int endingLineNumber = command.lineNumber_;

         StringExpr::skipSeparators(f,l);

         if(f == l)
         {
             messageFiles.writeBytes(2, ErrorMessageWithLocation("block-statement:  missing end of block",
                                                                 endingLineStart,
                                                                 command
                                                                ) + nl);
             return false;
         }

         if(*f != terminatingCharacter)
         {
             char buffer[2];
             buffer[0] = terminatingCharacter;
             buffer[1] = 0;

             messageFiles.writeBytes(2, errorKeyword + startFile + colon + lexical_cast<string>(startLine) + colon + 
                                        string(" incorrect terminator, found: ")
                                        + string(buffer) 
                                        + string("ending line was ") 
                                        + lexical_cast<string>(endingLineNumber) + nl);
             return false;
         }

         string postTerminatorFile = command.filename_;
         int    postTerminatorLine = command.lineNumber_;

         ++f;
         StringExpr::skipSeparators(f,l);

         
         pair<string,RedirType::REDIRECTION_LIST*> rd = RedirType::parseOnlyRedirections(f,l);

         if(rd.second)
             redirections.reset(rd.second);
         else
         {
            messageFiles.writeBytes(2, errorKeyword + postTerminatorFile + colon + lexical_cast<string>(postTerminatorLine)
                                       + colon + string("block-statement syntax error after ')' or '}' keyword") +nl);
            return false;
         }


         string::const_iterator saveF = f;

         while(f != l)
         {
             saveF = f;

             auto_ptr<StringExpression> token( StringExpr::parse(f, l) );

             if(!token.get())
                break;
             
             if(token->isSpecial())  // propably the end of token
             {
                 string text = token->stringify();

                 if(text == ";;")
                 {
                    f = saveF;
                    break;
                 }

                 if(text[0] == ';')
                 {
                    break;
                 }

                 f = saveF;
                 break;
             }
             
             StringExpr::skipSeparators(f,l);
         }

         break;  // let the end of function handle the insertion of the statement list into the output data
      }

      StringExpr::skipSeparators(f,l);

      if(f == l)
      {
          if (!handleLineContinuations(stream, command))
          {
              if (stream.error())
                  return false;
              break;
          }

          f = command.text_.begin();
          l = command.text_.end();
      }

   }  // readstatements

   // now stick the statement list into a block statement along with the redirections

   // for exception saftey, do not separate the following lines
   ThreadSafeHandle<Statement>  listMembers(statementListStatement.get());
   statementListStatement.release();

   auto_ptr<BourneBlockStatement>   blockStatement(new BourneBlockStatement(listMembers));

   blockStatement->setRedirections(redirections.release());


   {   
       output.statements_->push_back(ThreadSafeHandle<Statement>(blockStatement.get()));
       blockStatement.release();
   }


   return true;
   
} // parse block statement

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseFunction      ( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               )
{
   // functions are defined in either of 2 ways:
   //
   //     "function"  functionName 
   //     {
   //        statements
   //     }
   //
   // Or
   //    functionName "()"
   //    {
   //       statements
   //    }
   // This function handles both.


   AutoPoppingPushPrompt  prompt(stream, stream.getPrompt() + ">" );  // will auto-restore on return
   

   string filename = stream.filename();
   int    lineNumber = stream.lineNumber();
   
   StringExpr::skipSeparators(f, l);
   
   if(f == l)
   {
       messageFiles.writeBytes(2, string("Error:  Attempting to parse a function at the end of file.\r\n"));
       return false;
   }

   string::const_iterator statementStart = f;
   
   string firstWord = parseFirstWord(f,l);

   size_t openParenOffset = firstWord.find_first_of('(');

   if ( openParenOffset < firstWord.size())
   {
       firstWord.erase(openParenOffset, firstWord.size()-1);
       f = statementStart + openParenOffset;
   } 

   bool functionKeywordFound = false;

   if(firstWord == "function")
   {
       functionKeywordFound = true;

       firstWord = parseFirstWord(f,l);

       if (firstWord.empty())
       {
           messageFiles.writeBytes(2, ErrorMessageWithLocation("function-syntax:  illegal name", statementStart, command) + nl); 
           return false;
       }

       if(firstWord.find_first_of("(){}") < firstWord.size())
       {
           messageFiles.writeBytes(2, ErrorMessageWithLocation("function-syntax:  illegal name", statementStart, command) + nl); 
           return false;
       }


   }

   if(!isFuncName(firstWord))
   {
       messageFiles.writeBytes(2, ErrorMessageWithLocation("function-syntax:  illegal name", statementStart, command) + nl); 
       return false;
   }

   StringExpr::skipSeparators(f,l);

   if(f == l)
   {
         string filename = command.filename_;

         if(!handleLineContinuations(stream, command))
         {
           messageFiles.writeBytes(2, string("Error:  End of file in function ") + firstWord + nl);
           return false;
         }

         f = command.text_.begin();
         l = command.text_.end();

         
         if(f == l)
         {
           messageFiles.writeBytes(2, string("Error:  Missing body for function ") + firstWord + in + filename + nl);
           return false;
         }

         statementStart = f;

   }

   if(   !functionKeywordFound 
      && f != l 
      && *f == '('
     )
   {
      parseFirstWord(f,l);  // eat the ()
   }
   
   StringExpr::skipSeparators(f, l);

   if (f == l)
   {
       string filename = command.filename_;

       if (!handleLineContinuations(stream, command))
       {
           messageFiles.writeBytes(2, string("Error:  End of file in function ") + firstWord + nl);
           return false;
       }

       f = command.text_.begin();
       l = command.text_.end();


       if (f == l)
       {
           messageFiles.writeBytes(2, string("Error:  Missing body for function ") + firstWord + in + filename + nl);
           return false;
       }

       statementStart = f;

   }

   while (f != l && *f == '#')
   {   // ignore comments after the function delcaraton and before the 
       f = l;

       if (!handleLineContinuations(stream, command))
       {
           messageFiles.writeBytes(2, string("Error:  End of file in function ") + firstWord + nl);
           return false;
       }

       f = command.text_.begin();
       l = command.text_.end();


       if (f == l)
       {
           messageFiles.writeBytes(2, string("Error:  Missing body for function ") + firstWord + in + filename + nl);
           return false;
       }

       StringExpr::skipSeparators(f, l);

       statementStart = f;
   }

   if(*f != '(' && *f != '{')
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("Function bodies must be defined using () or {}. ", statementStart, command) + nl );
        return false;
   }


   size_t oldSize = output.statements_->size();

   statementStart = f;


   EndOfStatementListDetector endDetector;  // the parseStatement will know to switch from "end of file" to ), or }, depending.

   ParseStatementReturnCode rc =parseStatement(
                                                f,
                                                l,
                                                stream,
                                                output,
                                                messageFiles,
                                                command,
                                                endDetector,
                                                parsingContext
                                               );

   if(rc != PSRC_GOOD)
      return false;

   if(output.statements_->size() <= oldSize)
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("ParseStatement failure", statementStart, command) + nl );
        return false;
      
   }

   ThreadSafeHandle<Statement> body = output.statements_->back();

   output.statements_->pop_back();

   auto_ptr<Statement> s(new InstallFunctionStatement(firstWord, body, filename, lineNumber));

   {
       // keep together for exception safety
       output.statements_->push_back( ThreadSafeHandle<Statement>(s.get()) );
       s.release();
   }


   return true;

}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



static bool  parseCaseMember(string::const_iterator           &f,
                             string::const_iterator           &l,
                             ScriptLineStream                 &stream,
                             ScriptThread::ManagedFileHandles &messageFiles,
                             ScriptInputLine                  &command,
                             CaseMemberList                   &outputMembers,
                             ScriptThread::ThreadContext      *parsingContext,
                             bool                             &parseError
                           )
{
    // a case member looks like this:
    //
    //    expression ) statement list ;;

    parseError = true;  // unless we leave successfully, the default state is error

    //-----------------------------------------------------------------------------
    //-----------------------------------------------------------------------------
    //-----------------------------------------------------------------------------
    // first, make sure that the header for the case member is correct.  Hears
    // can look like this:
    //
    //      value)
    //      value1 | value2 | value3 )
    //      (value)
    //      (value1|value2)


   StringExpr::skipSeparators(f, l);


    auto_ptr<StatementList>           statementList(new StatementList);
    auto_ptr<StatementListStatement>  statementListStatement;

    {
       // keep together for exception saftey

       statementListStatement.reset( new StatementListStatement(statementList.get()));
       statementList.release();
    }

    // prepare to parse something like this:
    //
    //    [(]  word [ | word ] ... ) statements ;;

    StringExpr::skipSeparators(f, l);

    // skip blank lines

    while (f == l)
    {
        if (!handleLineContinuations(stream, command))
        {
            if (stream.error())
                return false;
            break;
        }

        f = command.text_.begin();
        l = command.text_.end();

        StringExpr::skipSeparators(f, l);
    }

    if(f != l && *f == ';')
    {
       ++f;  // just skip a leading ; which looks like a newline
    }

    if(f != l && *f == '(')
    {
       // skip leading (

       ++f;
       StringExpr::skipSeparators(f, l);
    }

    string::const_iterator statementStart = f;


    //-----------------------------------------------------------------------------
    //-----------------------------------------------------------------------------
    //-----------------------------------------------------------------------------
    // now parse the word list that defines the trigger conditions for this member
    //
    //    value [ | value] ... )



    auto_ptr< list<StringExpression*> > words( new list<StringExpression*>() );

    for(;;)
    {

        auto_ptr< StringExpression > word( cxxtls::StringExpr::parse(f,l) );
        
        if(!word.get())
        {

            messageFiles.writeBytes(2, ErrorMessageWithLocation("case-syntax: invalid member definitions", statementStart,command) + nl);
            return false;
        }
        
        {
           words->push_back(word.get());
           word.release();
        }

        StringExpr::skipSeparators(f, l);

        if(f != l && *f == '|')
        {
            ++f;

        }
        else
        {
           break;
        }

    }

    string esacValue;

    list<StringExpression*> &wordList = *words.get();
    StringExpression*        firstWord = *wordList.begin();

    if (firstWord->isSimple(esacValue) && esacValue == "esac")
    {
      parseError = false;
      return false;  // end of list
    }

    StringExpr::skipSeparators(f, l);

    if(f == l || *f != ')')
    {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("case-syntax: member definition requires ')'",statementStart,command) + nl);
        return false;
    }
    else
    if (f != l)
        ++f;

    // skip blank lines

    StringExpr::skipSeparators(f, l);

    if (f != l && *f == '#')
    {
        while (f != l && *f != '\n')
           ++f;
    }

    while(f == l)
    {
        if (!handleLineContinuations(stream, command))
        {
            if (stream.error())
                return false;
            break;
        }

        f = command.text_.begin();
        l = command.text_.end();

        StringExpr::skipSeparators(f, l);
    }

    //-----------------------------------------------------------------------------
    //-----------------------------------------------------------------------------
    //-----------------------------------------------------------------------------
    // now parse statements until we hit either ;; or esac


   EndOfStatementListDetector endDetector("esac");

   bool doneKeywordFound = false; // not "done" but rather ";;"
   bool esacKeywordfound = false;

   while(!stream.eof())
   { // parse a list of statements into the same case member

       if (f != l && *f == ';')
       {
           string::const_iterator next = f;
           ++next;
           if (next != l && *next == ';')
           {
               f = next+1;
               doneKeywordFound = true;

               StringExpr::skipSeparators(f, l);

               if (f != l && *f == '#')
               {
                   while (f != l && *f != '\n')
                       ++f;
               }

               break;
           }
       }

      string::const_iterator memberStart = f;

      ParseStatementReturnCode rc = parseStatement(f, l, stream, *statementListStatement, messageFiles, command, endDetector, parsingContext);

      if(rc == PSRC_BAD)
         return false;  // io error or syntax error
      else
      if(rc != PSRC_GOOD)
      {
         string word = parseFirstWord(f,l);

         if (word == "esac")
         {
             f = memberStart;    // let the outer parse, the case statement parse, re-parse the esac keyword
             parseError = false;
             break;
         }
         else
         {
             messageFiles.writeBytes(2, ErrorMessageWithLocation("case-member:  syntax error", memberStart,command) + nl);
         }
       

         return false;
         
      }


      StringExpr::skipSeparators(f,l);

      if(f == l)
      {
          if (!handleLineContinuations(stream, command))
          {
              if (stream.error())
                  return false;
              break;
          }

          f = command.text_.begin();
          l = command.text_.end();
      }

   }  // parse a list of statements into the same case member

   //--------------------------------------------------------------------------------------------------
   //--------------------------------------------------------------------------------------------------
   //--------------------------------------------------------------------------------------------------
   // push the parsed statement into the output list

   {

       // keep together for exception safetey

       ThreadSafeHandle<CaseMember> member( new CaseMember(words.get(), statementListStatement.get()));
       words.release();
       statementListStatement.release();
       
       outputMembers.push_back( member );
       
   }



   if ( f == l)
   {
       string filename = command.filename_;
       int    line = command.lineNumber_;

       if (!handleLineContinuations(stream, command))
       {
           if (stream.error())
               return false;
           else
           {

               messageFiles.writeBytes(2, string("Error: end of file in case statement at ") + filename + colon + lexical_cast<string>(line)+nl);
               return false;
           }
       }

       f = command.text_.begin();
       l = command.text_.end();
   }



   parseError = false;
   return true;
}



static bool parseCaseStatement ( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               )
{
   // case statements look like this:
   //
   //   case 
   //        Expression) statementlist ;;
   //        Expression) statementlist ;;
   //   esac
   //  


   AutoPoppingPushPrompt  prompt(stream, stream.getPrompt() + ">" );  // will auto-restore on return
   
   StringExpr::skipSeparators(f, l);
   
   if(f == l)
   {
       messageFiles.writeBytes(2, string("Error:  Attempting to parse a case-statement at the end of file:  ") + command.filename_ + nl);
       return false;
   }

   string::const_iterator statementStart = f;
   string  statementStartFile = command.filename_;
   int     statementStartLine  = command.lineNumber_;
   
   string firstWord = parseFirstWord(f,l);

   if(firstWord != "case")
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("case-statement:  syntax error", statementStart,command) + nl);
        return false;
   }

   statementStart = f;

   auto_ptr< StringExpression > word( cxxtls::StringExpr::parse(f,l) );

   if(!word.get())
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("case-syntax:  expected word after 'case'", statementStart,command) + nl);
        return false;
   }

   string::const_iterator beforeIn = f;

   string inWord = parseFirstWord(f,l);

   if(inWord != "in")
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("case-syntax:  expected 'in' after 'case word'",
                                   beforeIn, command) + nl );
        return false;
   }

   StringExpr::skipSeparators(f, l);

   if (f != l && *f == '#')
   {
       while (f != l && *f != '\n')
          ++f;
   }

   if (f == l)
   {
       string filename = command.filename_;
       int    line     = command.lineNumber_;

       if (!handleLineContinuations(stream, command))
       {
           if (stream.error())
               return false;
           else
           {

              messageFiles.writeBytes(2, string("Error: end of file in case statement at ") + filename + colon + lexical_cast<string>(line) + nl );
              return false; 
           }
       }

       f = command.text_.begin();
       l = command.text_.end();
   }



   auto_ptr< list< ThreadSafeHandle<CaseMember> > >  caseMembers(new list< ThreadSafeHandle<CaseMember> > );

   while(f != l)
   {
      bool parseError = false;

      statementStart = f;

      bool good = parseCaseMember(f, l, stream, messageFiles, command, *caseMembers, parsingContext, parseError);

      if (parseError)
      {
          messageFiles.writeBytes(2, ErrorMessageWithLocation("case-member-syntax", statementStart, command) + nl);
          return false; 
      }


      if(!good)
         break;  
   }

   if (caseMembers->empty())
   {
       messageFiles.writeBytes(2, errorKeyword + statementStartFile + colon + lexical_cast<string>(statementStartLine) 
                                  + colon + " case-syntax: no members" + nl );
       return false;
   }

   CXXTLS_FOREACH(ThreadSafeHandle<CaseMember> &cur, *caseMembers)
   {
      if(cur->conditions_ == 0)
      {
       messageFiles.writeBytes(2, errorKeyword + statementStartFile + colon + lexical_cast<string>(statementStartLine) 
                                  + colon + " case-statement-parse-error:  cases without trigger conditions" + nl );
       return false;
      }
   }


   auto_ptr< BourneCaseStatement > statement;

   {
      // keep together for exception safety
      statement.reset( new BourneCaseStatement(word.get(), caseMembers.get(), 0) );  // handle redirections later
      word.release();
      caseMembers.release();
   }

   {
      // keep together for exception saftey
      ThreadSafeHandle<Statement> rv( statement.get() );
      statement.release();
      output.statements_->push_back(rv);
   }


   return true;

}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseNotStatement  ( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 StatementListStatement           &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command,   // over which f and l iterate
                                 ScriptThread::ThreadContext      *parsingContext
                               )
{

   // the not statement looks like this:
   //
   //   ! statement

   AutoPoppingPushPrompt  prompt(stream, stream.getPrompt() + ">" );  // will auto-restore on return
   
   StringExpr::skipSeparators(f, l);
   
   if(f == l)
   {
       messageFiles.writeBytes(2, string("Error:  Attempting to parse a !statement (not-statement) at the end of file.\r\n"));
       return false;
   }

   string::const_iterator statementStart = f;
   

   if(*f != '!')
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("!-statement syntax error", statementStart, command) + nl);
        return false;
   }

   ++f;

   statementStart = f;

   EndOfStatementListDetector endDetector;

   ParseStatementReturnCode rc = parseStatement(f, l, stream, output, messageFiles, command, endDetector, parsingContext);

   if(rc != PSRC_GOOD)
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("!-statement-syntax error", statementStart, command) + nl);
        return false;
   }


   // at this point, we have a statement properly appended to the end of the statement list statement

   ThreadSafeHandle<Statement> mainClause;

   {
       // keep together for consistency
       mainClause = output.statements_->back();
       output.statements_->pop_back();
   }

   ThreadSafeHandle<Statement>  notStatement(new NotStatement(mainClause));

   output.statements_->push_back(notStatement);

   return true;


}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseArithmeticStatement  ( string::const_iterator           &f, 
                                         string::const_iterator           &l_in,
                                         ScriptLineStream                 &stream,
                                         StatementListStatement           &output,
                                         ScriptThread::ManagedFileHandles &messageFiles,
                                         ScriptInputLine                  &command   // over which f and l iterate
                                       )
{

   // the arithmetic statement looks like this:
   //
   //   ((  expression )) [ ; stuff ]
   //
   // and it acts like this:
   //
   //   let "expression"
   //
   //
   // which means we are parsing all the text between (( and )) and calling that a single, litteral token,
   // which we will foist off on a "let" command with only the one command line parameter.
   //
   // let statements are implemented as a builtin command which evaluates all its arguments
   // and treats each as an assignment or a comparison.

   std::string startFilename = stream.filename();
   int         startLineNumber = stream.lineNumber();


   AutoPoppingPushPrompt  prompt(stream, stream.getPrompt() + ">" );  // will auto-restore on return

   string::const_iterator l = l_in;

   {
      // find true end of the statement, which may be before l_in.  The true end would be a ; or }

      string::const_iterator scan = f;

      while(scan != l)
      {
          if(*scan == ';' || *scan == '}')
          {
             l = scan;
             break;
          }
          ++scan;
      }

   }

   // l has now been adjusted so as not to include statement separators


   StringExpr::skipSeparators(f, l);
   
   if(f == l)
   {
       messageFiles.writeBytes(2, string("Error:  Attempting to parse an arithmetic statement at the end of file.\r\n"));
       return false;
   }

   string::const_iterator statementStart = f;
   

   if(*f != '(')
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("((-statement syntax error",f, command) + nl);
        return false;
   }

   ++f; // skip the leading ((

   if(f == l)
   {
        messageFiles.writeBytes(2, ErrorMessageWithLocation("((-statement-syntax-error:  unexpected end of line", statementStart, command) + nl );
        return false;
      
   }

   ++f; // skip the leading ((

   StringExpr::skipSeparators(f, l);

   // find end of expression, which will be the last )) in the stream

   string::const_iterator tail = l;

   bool foundEnd = false;

   while(tail != f)
   {
       --tail;

       if (tail == f)
          break;

       if (*tail == ')')
       {
           --tail;

           if (*tail == ')')
           {
              // we have found the termination condition expected by the (( ... )) statement, but lets confirm that there is
              // nothing on the line after the ))


               foundEnd = true;

               string::const_iterator scan = tail + 2;

               StringExpr::skipSeparators(scan, l);

               if (scan != l)
               {
                   messageFiles.writeBytes(2, errorKeyword + command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + " '(('-statement:  invalid text after ))\r\n");
                   return false;
               }


              l = tail;   // end of input region, (f,l], defines the expression of interest

              break;   
           }
           // the final ) on the line is not preceeded by )), so we have a syntax error

           messageFiles.writeBytes(2, errorKeyword + command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + " '(('-statement is missing trailing ))\r\n");
           return false;
       }

       
   }

   if (!foundEnd)
   {
       messageFiles.writeBytes(2, errorKeyword + command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + " '(('-statement is missing trailing ))\r\n");
       return false;
   }

   // now we are going to copy (f,l] to the expression variable

   std::string expression;

   statementStart = f;

   expression.assign(f,l);
   
   f = l+2;   // need to set this assignment here because the caller of this function may check f to ensure that there is no junk after
              // the valid arithmetic expression part.  f is left after the )) that defines the expression.  If no )) exists, then an
              // error would have occurred earlier.

   if(expression.find_first_of("'\"") < expression.size())
   {
        messageFiles.writeBytes(2, errorKeyword + command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + " '(('-statement contained apostrophe or double quote\r\n");
        return false;
   }



   std::string::const_iterator beg = expression.begin(),
                               end = expression.end();

#if 0
   auto_ptr<StringExpr::Arithmetic::Node> checker(StringExpr::Arithmetic::parseExpression(beg, end));

   if(checker.get() == 0)
   {
        messageFiles.writeBytes(2, errorKeyword + command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + colon + "'(('-statement syntax error\n");
        return false;
   }
#endif

   // construct a let statement given only the one expression and append it to the output

   auto_ptr<BrashCommandLine::ARGUMENTS_LIST  >       arguments    (new BrashCommandLine::ARGUMENTS_LIST);
   auto_ptr<BrashCommandLine::REDIRECTION_LIST>       redirections (new BrashCommandLine::REDIRECTION_LIST);

   auto_ptr< BrashCommandLine::SUPERSEEDS >  superseeds(new BrashCommandLine::SUPERSEEDS);


   string::const_iterator letBeg = let.begin(),
                          letEnd = let.end();

   auto_ptr<cxxtls::StringExpression>   letExpression( cxxtls::StringExpr::parse(letBeg, letEnd) ); // the word "let"

#if 0
   auto_ptr<cxxtls::StringExpression>   valueExpression( new StringExpression );                    // the text between (( and ))

   auto_ptr<StringExpr::LiteralNode>    expressionNode( new StringExpr::LiteralNode(expression) );

   {
       // keep together for exception safety
       valueExpression->adoptNode( expressionNode.get() );
       expressionNode.release();
   }

#else

   string decoratedExpression("\"");  

   decoratedExpression += expression;

   decoratedExpression += "\"";

   string::const_iterator parseStart = decoratedExpression.begin(),
                          parseEnd   = decoratedExpression.end();

   auto_ptr<cxxtls::StringExpression>   valueExpression( cxxtls::StringExpr::parseStrExpr(parseStart, parseEnd, false) );  // ignore > and <

   if(valueExpression.get() == 0)
   {
        messageFiles.writeBytes(2, errorKeyword + command.filename_ + colon + lexical_cast<string>(command.lineNumber_) + " '(('-statement:  invalid expression syntax\r\n");
        return false;
   }
#endif

   auto_ptr<BrashCommandLine> commandLine;

   {
      // for exception saftey, do not separate the following lines

      arguments->push_back(letExpression.get());
      letExpression.release();

      arguments->push_back(valueExpression.get());
      valueExpression.release();
   }


   {
       //for exception saftey, do not separate the following lines

       commandLine.reset( new BrashCommandLine(arguments.get(), redirections.get(), superseeds.get(), startFilename, startLineNumber) );
       arguments.release();
       redirections.release();
       superseeds.release();
   }

   {
      // for exception saftey, do not separate the following lines

      bool maybeBuiltin = true; // actually, its not, but the truth doesn't matter in this case
      
      output.statements_->push_back(ThreadSafeHandle<Statement>(new BrashCommandLineStatement(commandLine.get(), maybeBuiltin)));
      commandLine.release();
   }

   return true; 

}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool parseFcStatement    ( string::const_iterator           &f, 
                                   string::const_iterator           &l,
                                   ScriptLineStream                 &stream,
                                   StatementListStatement           &output,
                                   ScriptThread::ManagedFileHandles &messageFiles,
                                   ScriptInputLine                  &command,   // over which f and l iterate
                                   ScriptThread::ThreadContext      *parsingContext
                                 )
    /// parse and execute, at parse time, the fc command behavior.
try
{

    std::vector<string> const &history = stream.history();

    vector< ThreadSafeHandle<StringExpression> >  commandLineExpressions;

    // parse the command line into a vector of string expressions up to the trailing \n or ; 

    while(f != l)
    {
        StringExpr::skipSeparators(f,l);

        if(f == l)
        {
           break;
        }

        if(*f == ';')
        {
           string::const_iterator next = f; ++next;

           if(next == l)
               break;
           
           if(*next == ';')
              break;
        }

        auto_ptr<StringExpression>  expr(StringExpr::parseStrExpr(f,l, false));

        if(!expr.get())
        {
            messageFiles.writeBytes(1, string("bash:  fc:  invalid expression syntax\r\n"));
            return true;
        }

        {
           // keep together for exception saftey

           ThreadSafeHandle<StringExpression> tmp(expr.get());
           expr.release();
           commandLineExpressions.push_back(tmp);
        }

    }

    if(commandLineExpressions.size() < 2)
    {
       messageFiles.writeBytes(1, string("bash:  fc: invalid invocation.  Use -h for help.\r\n"));
       return true;
    }

    // now execute the string expressions to create an array of strings for use as command line
    // arguments for the code below

    vector<string> parms;

    CXXTLS_FOREACH(ThreadSafeHandle<StringExpression> &cur, commandLineExpressions)
    {
        string value;

        StringArray output;

        cur->execute(&parsingContext->variables(), output);

        value = output.combine();

        parms.push_back(value);

    }

    vector<char*>  args;  // paralles parms

    CXXTLS_FOREACH(string const &cur, parms)
        args.push_back( const_cast<char*>(cur.c_str()) );

    // args is now a command line array just like main() gets


    int argc = 1;

    // figure out how many argc values we should lie and say are in the argv.
    // 
    // this stupid command has this kind of options
    //
    //   fc -l -10 -20 
    //
    // Note that -10 looks like an option, but it is not, it is a value
    //
    // the fc command is stupid in this regard, but hey we are stuck with it

    while(argc < int(args.size()))
    {
        if(args[argc][0] != '-')
           break;

        if(parms[argc].size() < 2)
           break;
        
        if (isdigit(args[argc][1]))  // stop on first -argument with a digit as the option character
        {
           break;
        }


        ++argc;
    }


    char ** argv = &args[0];

    GetOpts optProc(argc, const_cast<char *const *&>(argv), "hslrn");
                                                             
    int c;

    // now interpret the command line options

    bool listOption = false;
    bool submitOption = false;

    int  firstItem = 0;
    int  lastItem  = 0;

    bool numbers   = true;

    bool done = false;

    while( !done && (c = optProc.getNextOption() ) != -1 )
    switch(c)
    {
        case 'h':
            {
                messageFiles.writeBytes(1, string("\r\n\r\nfc:  fix commands commands previously entered.\nUsage:\r\n"));
                messageFiles.writeBytes(1, string("    1) fc [firstCommand] [lastCommand]:  edit and resubmit previous commands\r\n\r\n"));
                messageFiles.writeBytes(1, string("    2) fc [-lrn] [firstCommand] [lastcommand]:  list previous commands by number.  \r\n"));
                messageFiles.writeBytes(1, string("                                                -n suppresses numbering\r\n\r\n"));
                messageFiles.writeBytes(1, string("    3) fc -s [str1=str2] [commandNumber]:  resubmit a command with string substitition\r\n\r\n"));
                messageFiles.writeBytes(1, string("    4) fc -h:  get help\n\r\n"));
                messageFiles.writeBytes(1, string("Description:\n\r\n"));
                messageFiles.writeBytes(1, string("    The following environment variables are searched to determine your editor's\r\n"));
                messageFiles.writeBytes(1, string("    command name:\r\n"));
                messageFiles.writeBytes(1, string("        FCEDIT\r\n"));
                messageFiles.writeBytes(1, string("        EDITOR\r\n"));
                messageFiles.writeBytes(1, string("        VISUAL_EDITOR\r\n"));
                
                messageFiles.writeBytes(1, string("\r\n")); // the end

                return true;
            }
            break;
        case 's':
            submitOption = true;
            break;
        
        case 'l':  
            listOption = true;
            break;

        case 'r':  
            {
                messageFiles.writeBytes(2, string("brash:  fc:   -r option not supported\r\n"));
                return true;
            }
            break;

        case 'n':
            {
               numbers = false;
            }
            break;

        default:
                if(    optProc.optind >= int(args.size()) 
                    || 0 == strlen(args[optProc.optind])
                  )
                    break;

                if( isdigit(args[optProc.optind][1]) )
                {
                   done = true;
                   break;
                }

                messageFiles.writeBytes(2, string("brash:  fc:   unknown option\r\n"));
                return true;
            
    }


    /////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////

    if(submitOption)
    {
       // we are expecting syntax like this:
       //
       //   fc -s [str1=str2] command
       //
       // We should have alread parsed the -s option, and should now be looking at str1=str2

       string oldString, newString;

       if(args.size() < 3)
       {
           messageFiles.writeBytes(2, string("brash: fc:  -s option missing needed parameters\r\n"));
           return true;
       }

       if(args.size() > 4)
       {
           messageFiles.writeBytes(2, string("brash: fc:  -s option: too many parameters\r\n"));
           return true;
       }

       if(args.size() == 4)
       {
          string substArgs = args[2];

          size_t equal = substArgs.find_first_of('=');

          if(equal >= substArgs.size())
          {
              messageFiles.writeBytes(2, string("brash: fc -s: expected:  -s [str1=str2] commandNumber\n"));
              return true;
          }

          string substarg  = args[2];

          oldString=substarg.substr(0,equal);
          newString=substarg.substr(equal+1);

          string commandarg = args[3];

          if( commandarg.empty() || !isdigit(commandarg[0]) )
          {
              messageFiles.writeBytes(2, string("brash: fc -s: invalid:  commandNumber\n"));
              return true;
          }

          int commandNumber = lexical_cast<int>(args[3]);

          if(commandNumber < 1 || commandNumber > history.size())
          {
              messageFiles.writeBytes(2, string("brash: fc -s: command Number out of range\r\n"));
              return true;
          }

          vector<string>::const_iterator scan = history.begin();

          int count=0;

          --commandNumber;  // switch to 0 based

          while(count < commandNumber)
          {
              ++count;
              ++scan;
          }

          string commandValue = *scan;

          vector<string> nonMatchingPieces;

          StrTool::splitOnString(commandValue, nonMatchingPieces, oldString);

          string newCommandLine;

          vector<string>::const_iterator f = nonMatchingPieces.begin(),
                                         l = nonMatchingPieces.end();

          while(f != l)
          {
             string const &piece = *f++;

             newCommandLine += piece;
             
             if(f != l)
             {
                newCommandLine += newString;
             }

          }

          // we now have a winner.

          newCommandLine;


          stream.addToHistory(newCommandLine);

          string tmpdir;

          if(parsingContext->variables().getTmpDirName(tmpdir))
          {
             messageFiles.writeBytes(1, string("brash:  fc:  cannot find TMP directory\r\n"));
             return true;
          }

          string tmpfilename = tmpdir + "/fc_subst_run.txt";


          {
              ofstream outfile(tmpfilename.c_str());
              
              if(!outfile.good())
              {
                 messageFiles.writeBytes(1, string("brash:  fc:  cannot write tmp file: ") + tmpfilename + nl);
                 return true;
              }

              outfile << newCommandLine << endl;
          }

          {
              cxxtls::StringExpr::ExecutionContext::ArgsVector    args;

              args.push_back(tmpfilename);

              brashReadAndExecuteInContext(tmpfilename, parsingContext, args, false);
          }
  
          FileName deleter(tmpfilename);
  
          deleter.remove();


          return true;  // we are done here

       }
       else
       {
          // args.size == 3, and we are just going rerun the specified command

          string commandarg = args[2];

          if( commandarg.empty() || !isdigit(commandarg[0]) )
          {
              messageFiles.writeBytes(2, string("brash: fc -s: invalid:  commandNumber\n"));
              return true;
          }

          int commandNumber = lexical_cast<int>(commandarg);

          if(commandNumber < 1 || commandNumber > history.size())
          {
              messageFiles.writeBytes(2, string("brash: fc -s: command Number out of range\r\n"));
              return true;
          }

          vector<string>::const_iterator scan = history.begin();

          int count=0;

          --commandNumber;  // switch to 0 based

          while(count < commandNumber)
          {
              ++count;
              ++scan;
          }

          string commandValue = *scan;

          string tmpdir;

          if(parsingContext->variables().getTmpDirName(tmpdir))
          {
             messageFiles.writeBytes(1, string("brash:  fc:  cannot find TMP directory\r\n"));
             return true;
          }
       
          string tmpfilename = tmpdir + "/fc_subst_run.txt";


          {
              ofstream outfile(tmpfilename.c_str());
              
              if(!outfile.good())
              {
                 messageFiles.writeBytes(1, string("brash:  fc:  cannot write tmp file: ") + tmpfilename + nl);
                 return true;
              }

              outfile << commandValue << endl;
          }

          {
              cxxtls::StringExpr::ExecutionContext::ArgsVector    args;

              args.push_back(tmpfilename);
              
              brashReadAndExecuteInContext(tmpfilename, parsingContext, args, false);
          }
  
          FileName deleter(tmpfilename);
  
          deleter.remove();


          return true;  // we are done here


       }


    }

    /////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////


    int count=0;

    for (int index = optProc.optind; index < int(args.size()); index++)
    {
       ++count;

       if(count == 1)
          firstItem = lexical_cast<int>(args[index]);
       else
       if(count == 2)
          lastItem = lexical_cast<int>(args[index]);
       else
       {
          messageFiles.writeBytes(2, string("brash:  fc: too many non-option parameters \r\n"));
          return true;
       }
    }

    if(!listOption && firstItem == 0)
    {
       firstItem = lastItem = history.size();
    }


    if(firstItem == 0)
    {
       // firstItem is not specified, start 16 back from the end

       firstItem = int(history.size()) - 16;

       if(firstItem < 1)
          firstItem = 1;
    }
    else
    if(firstItem < 0)
    {
        firstItem = int(history.size()) + firstItem;

        if(firstItem < 1)
           firstItem = 1;
    }
    else
    if (firstItem > int(history.size()))
    {
        return true; // nothing to show
    }


    if(lastItem == 0)
    {
       // last item not specified

       lastItem = firstItem + 16;
    }
    else
    if(lastItem < 0)
    {
       lastItem = int(history.size()) + lastItem;

       if(lastItem < firstItem)
          lastItem = firstItem;
    }
    
    if(lastItem >= int(history.size()))
       lastItem = history.size();       // first and last item are 1 based, not zero based

    if (firstItem < 1)
        firstItem = 1;

    if (lastItem < 1)
       lastItem = 1;

    if(lastItem - firstItem > 1000)
    {
       messageFiles.writeBytes(1, string("brash:  fc:  requested range is too large (>1000 lines)\r\n"));
       return true;
    }


    if(listOption)
    {
    /////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////
        int row=0;

        if (history.empty())
        {
            messageFiles.writeBytes(1, string("history is empty\r\n"));
        }
        
        CXXTLS_FOREACH(string const &cur, history)
        {
           ++row;
        
           if(row >= firstItem && row <= lastItem)
           {
               if(numbers)
                   messageFiles.writeBytes(1, lexical_cast<string>(row) + space);
        
               messageFiles.writeBytes(1, cur + nl);
           }
           
        }
    }
    else
    {
    /////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////
        if(history.empty())
        {
           messageFiles.writeBytes(1, string("brash:  fc:  history is empty\r\n"));
           return true;
        }


        string tmpdir = "/tmp";

        if(parsingContext->variables().getTmpDirName(tmpdir))
        {
           messageFiles.writeBytes(2, string("brash:  fc:  cannot determine TMP directory name\r\n"));
           return true;
        }

        string editor;

        editor = parsingContext->variables().variables().get("FCEDIT");

        if(editor.empty())
        {
            editor = parsingContext->variables().variables().get("EDITOR");
        }

        if(editor.empty())
        {
            editor = parsingContext->variables().variables().get("VISUAL_EDITOR");
        }


        if(editor.empty())
        {
            editor = parsingContext->variables().variables().get("VISUAL");
        }

        if(editor.empty())
        {
            editor="edit";
        }

        FileName tmpEditor(editor);

        FileName fullname = tmpEditor.find_executable(editor, parsingContext->variables().variables().get("PATH"));

        if(fullname.empty() || !fullname.exists())
        {
           messageFiles.writeBytes(1, string("brash:  fc:  cannot find editor: ") + editor + nl);
           return true;
        }

        string tmpfilename(tmpdir + "/" + "fcedit.txt");

        {
            // populate the temp file and let us edit it.

            ofstream tmpFile(tmpfilename);
            
            if(!tmpFile.good())
            {
               messageFiles.writeBytes(1, string("brash:  fc:  cannot create tmpfile ") + tmpfilename + nl);
               return true;
            }

            stream.clearGroup();

            
            int row = 0;
            CXXTLS_FOREACH(string const &cur, history)
            {
               ++row;
            
               if(row >= firstItem && row <= lastItem)
               {
                   stream.addToHistory(cur);
                   tmpFile << cur << endl;
               }
               
            }

        }

        string editCommand = fullname + space +  tmpfilename;

        system(editCommand.c_str());

        messageFiles.writeBytes(1, nl);


        {
            cxxtls::StringExpr::ExecutionContext::ArgsVector    args;

            args.push_back(tmpfilename);

            brashReadAndExecuteInContext(tmpfilename, parsingContext, args, false);
        }

        FileName deleter(tmpfilename);

        deleter.remove();
        

    }

    return true;
}
catch(...)
{
    messageFiles.writeBytes(1, string("brash:  fc command:  bad parameters exception.\r\n"));
    return true;
}

