#ifndef cxxtls_pseudoterminal_h
#define cxxtls_pseudoterminal_h
#ifdef __linux__

// LINUX AND UNIX ONLY!  Maybe ming but who knows.  On Ming/windows this may not be
// necessary for the primary reason that I created the file.  My primary reason was
// so that I could control gdb in order to write my own front end for it.

//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

///@file
///    This file declares the class, cxxtls::PseudoTerminal
///

#include <string>
#include <cxxtls/thread.h>

namespace cxxtls
{

class PseudoTerminal
    ///  A PseudoTerminal is a wrapper around the unix/linux "pty" concept.
    ///
    ///  For example programs using PseudoTerminals, look at:
    ///
    ///  @li tests/test_pty[12].cxx
    ///  @li bin/q.cxx
    ///
    ///  A Pty is a pseudo terminal -- a software emulation of the h/w "terminal"
    ///  concept.  Basically, anything that a file handle to a real terminal can do,
    ///  a file handle to a pseudo terminal can do.  The xterm application uses a 
    ///  pseudo terminal to host the shell that it runs in.
    ///
    ///  A pty (and thus a PseudoTerminal) is meant to be used as a controller for a 
    ///  child process which will both read and write to the PTY.  Note that it is important
    ///  that the child process have a file opened for write to the PTY because if it does not
    ///  bad things can happen when you close the pseudo terminal.  It technically works but 
    ///  there are serious annoyances if you don't.  See note 1.
    ///
    ///  A PseudoTerminal is actually a pair of interfaces that are tied together such
    ///  that what is written to one can be read from the other.  This allows two programs,
    ///  each acting as if it is interacting with a terminal, to be actually interacting 
    ///  with each other.  The two terminals are referred to as "master" and "slave".  A
    ///  program creating the pty, has programatic control of the slave and the assumption
    ///  is that some other program will be reading and writing to the slave.  
    ///
    ///  Threads, while causing difficulties in their own right, can make interaction with
    ///  pseudo terminals much simpler because the code reading from the terminals won't have
    ///  to perform selects, timeouts, etc that are required without them.  Of course, using
    ///  threads means that you have to mutex lock access to stdout, cout, cerr, etc, so they
    ///  cause their own problems.
    ///
    ///  User logins for example make heavy use of terminal properties -- even though you
    ///  may not be using a real terminal to log into.
    ///
    ///  There is typically a limit of 2048 total pseudo terminals on any given host, so
    ///  don't waste them.
    ///
    ///  See http://rachid.koucha.free.fr/tech_corner/pty_pdip.html for a good overview
    ///  of how to use pseudo terminals and how they work.
    ///
    ///  Generally though, you can think of psuedo terminals as a way of faking a terminal in
    ///  memory.  This is different than simply using a normal file handle because terminals have 
    ///  very complicated functionality.  See <termios.h>.   Also @see terminalstatus.h
    ///
    ///  When using gdb to debug a curses application, it is helpful to have a pseudo terminal
    ///  that used to actually handle the curses emulation.  That way, gdb does not have to
    ///  completely simulate the logic of curses in order to be useful.  Gdb has an option, -tty,
    ///  that lets you give then name of a slave terminal.
    ///
    ///  To use this class follow this example:
    ///
    ///@code
    ///
    ///   PsuedoTerminal pt;   // the state is currently "unopened"
    ///
    ///   pt.open(false);      // if successfull the state is now good (open canonical)
    ///
    ///   if(pt.state() != PseudoTerminal::good)
    ///   {
    ///       cout << "Error: PseudoTerminal error:  " << pt.stateNameString() << endl; 
    ///   }
    ///
    ///   string command = "echo hello world >";
    ///
    ///   command += pt.slaveTerminalName();
    ///
    ///   system(command.c_str());  // write to the slave and thus make data
    ///                             // available to the master.
    ///
    ///   char buffer[1024];
    ///
    ///   int rc = read(pt.masterFD(), buffer, sizeof(buffer));
    ///
    ///   if(rc < 0)
    ///   {
    ///       cout << "Error:  Data not flowing through PsuedoTerminal" << endl;
    ///   }
    ///   else
    ///   {
    ///      if(rc > 0)  
    ///          buffer[rc] = 0; // null terminate string read from pty
    ///      else
    ///          buffer[0] = 0;
    ///
    ///      cout << "You should see hello world here:  " << buffer << endl;
    ///   }
    ///
    ///@endcode
    ///
    /// @b Notes:
    ///
    /// 1.  Problem getting end of file if there is a process ONLY reading from the slave device
    ///
    ///     If you have a process that is ONLY reading from the slave device, and if you have
    ///     a thread reading from the masterFD, and if have a different thread close the masterFD,
    ///     the thread reading from the masterFD will NOT RECEIVE AN END OF FILE.  This means that
    ///     merely closing closing the masterFD or deleting the pty all together will not terminate
    ///     your reader thread.  You must manually track down the offending read-only process 
    ///     (such as 'cat &lt;/dev/pts/1') and kill the process yourself.  This is not that great of a
    ///     difficulty if you have created the child process interacting with the save pty.  But if
    ///     someone else grabs your PTY device, while it is open, they can cause your reader thread to hang.
    ///
    ///     Note that if the external process that is reading the slave also writes to the slave, it will
    ///     receive a SIGHUP notification, and SHOULD DIE (hopefully it will anyway -- cat will).
    ///     There is not much you can do about this.  Its basically a deadlock caused by the way the 
    ///     pty is being used -- just don't do this.   
    /// 

{
public:

    PseudoTerminal(); 
   ~PseudoTerminal();

    int masterFD() const
        /// Return the file handle to the master terminal.
        ///
        /// That which is written to the master terminal will become available
        /// to the slave terminal to read and vice versa.
        /// 
    {
         ScopedMutexLock lock( const_cast<Mutex&>(mutex_) );

         return masterFD_;
    }

    std::string const &slaveTerminalName() const
        /// Return the name of the slave terminal.  It can then be opened like
        /// any other terminal device can be opened.  You can read data written
        /// to the master terminal.  Anything written to the slave can be read
        /// by the master.
    {
        ScopedMutexLock lock(  const_cast<Mutex&>(mutex_)  );

        return slaveTerminalName_;
    }

    enum StateValue
    {
        good,                  ///< opened and all is well
        unopenable,            ///< openfailed
        unattributable,        ///< open ok but something else wrong
        ungrantable,           ///< open ok but something else wrong
        locked,                ///< open ok but something else wrong
        unnameable,            ///< open ok but somethign else wrong
        unopened,              ///< no open attempt yet (or closed)
        
        COUNT // must be last
    };

    StateValue state() const
         ///  Get the current state
    {
        ScopedMutexLock lock( const_cast<Mutex&>(mutex_) );

        return state_;
    }

    std::string stateNameString() const;
        ///<  Get the current state of the terminal as a string

    StateValue open(bool raw);  // true means raw i/o, otherwise canonical
    void       close();

    void setWindowSize(int rows, int columns);  ///<  Probably useless, but it
                                                ///   makes the termios calls to
                                                ///   tell the slave what the height
                                                ///   and width of it's window is.
                                                ///   In case the program using the
                                                ///   slave can format text to fit
                                                ///   the window more precisely.

private:

    Mutex       mutex_;

    int         masterFD_;
    StateValue  state_;

    std::string slaveTerminalName_;

    bool        raw_;     // raw or canonical i/o


}; // class PseudoTerminal

} // namespace cxxtls


#endif // not _MSC_VER
#endif // cxxtls_pseudoterminal_h
